# Copyright 2013 - Mirantis, Inc.
# Copyright 2015 - StackStorm, Inc.
# Copyright 2015 Huawei Technologies Co., Ltd.
# Copyright 2016 - Brocade Communications Systems, Inc.
# Copyright 2018 - Extreme Networks, Inc.
# Copyright 2019 - NetCracker Technology Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from oslo_log import log as logging
from oslo_utils import uuidutils
from pecan import rest
from wsme import types as wtypes
import wsmeext.pecan as wsme_pecan

from mistral.api import access_control as acl
from mistral.api.controllers.v2 import execution_report
from mistral.api.controllers.v2 import resources
from mistral.api.controllers.v2 import sub_execution
from mistral.api.controllers.v2 import task
from mistral.api.controllers.v2 import types
from mistral import context
from mistral.db.v2 import api as db_api
from mistral.db.v2.sqlalchemy import models as db_models
from mistral import exceptions as exc
from mistral.rpc import clients as rpc
from mistral.services import workflows as wf_service
from mistral.utils import filter_utils
from mistral.utils import rest_utils
from mistral.workflow import data_flow
from mistral.workflow import states
from mistral_lib.utils import merge_dicts

LOG = logging.getLogger(__name__)

STATE_TYPES = wtypes.Enum(
    str,
    states.IDLE,
    states.RUNNING,
    states.SUCCESS,
    states.ERROR,
    states.PAUSED,
    states.CANCELLED
)


def _get_workflow_execution_resource_with_output(wf_ex):
    rest_utils.load_deferred_fields(wf_ex, ['params', 'input', 'output'])

    return resources.Execution.from_db_model(wf_ex)


def _get_workflow_execution_resource(wf_ex):
    rest_utils.load_deferred_fields(wf_ex, ['params', 'input'])

    return resources.Execution.from_db_model(wf_ex)


# Use retries to prevent possible failures.
@rest_utils.rest_retry_on_db_error
def _get_workflow_execution(id, must_exist=True, fields=None):
    if fields and 'id' not in fields:
        fields.insert(0, 'id')

    fields_tuple = rest_utils.fields_list_to_cls_fields_tuple(
        db_models.WorkflowExecution,
        fields
    )

    with db_api.transaction():
        if must_exist:
            wf_ex = db_api.get_workflow_execution(id, fields=fields_tuple)
        else:
            wf_ex = db_api.load_workflow_execution(id, fields=fields_tuple)

        return rest_utils.load_deferred_fields(
            wf_ex,
            ['params', 'input', 'output', 'context', 'spec']
        )


# TODO(rakhmerov): Make sure to make all needed renaming on public API.


class ExecutionsController(rest.RestController):
    tasks = task.ExecutionTasksController()
    report = execution_report.ExecutionReportController()
    executions = sub_execution.SubExecutionsController()

    @rest_utils.wrap_wsme_controller_exception
    @wsme_pecan.wsexpose(resources.Execution, wtypes.text, types.uniquelist)
    def get(self, id, fields=None):
        """Return the specified Execution.

        :param id: UUID of execution to retrieve.
        :param fields: Optional. A specified list of fields of the resource to
                       be returned. 'id' will be included automatically in
                       fields if it's not provided.
        """
        acl.enforce("executions:get", context.ctx())

        LOG.debug("Fetch execution [id=%s]", id)

        wf_ex = _get_workflow_execution(id, fields=fields)

        if fields:
            return resources.Execution.from_tuples(zip(fields, wf_ex))

        resource = resources.Execution.from_db_model(wf_ex, fields=fields)

        resource.published_global = (
            data_flow.get_workflow_execution_published_global(wf_ex)
        )

        return resource

    @rest_utils.wrap_wsme_controller_exception
    @wsme_pecan.wsexpose(
        resources.Execution,
        wtypes.text,
        body=resources.Execution
    )
    def put(self, id, wf_ex):
        """Update the specified workflow execution.

        :param id: UUID of execution to update.
        :param wf_ex: Execution object.
        """
        acl.enforce('executions:update', context.ctx())

        LOG.info('Update execution [id=%s, execution=%s]', id, wf_ex)

        @rest_utils.rest_retry_on_db_error
        def _compute_delta(wf_ex):
            with db_api.transaction():
                # ensure that workflow execution exists
                wf_ex_old = db_api.get_workflow_execution(
                    id,
                    fields=(db_models.WorkflowExecution.id,
                            db_models.WorkflowExecution.root_execution_id)
                )
                root_execution_id = wf_ex_old.root_execution_id
                if not root_execution_id:
                    root_execution_id = wf_ex_old.id

                context.ctx(root_execution_id=root_execution_id)

                delta = {}

                if wf_ex.state:
                    delta['state'] = wf_ex.state

                if wf_ex.description:
                    delta['description'] = wf_ex.description

                if wf_ex.params and wf_ex.params.get('env'):
                    delta['env'] = wf_ex.params.get('env')

                # Currently we can change only state, description, or env.
                if len(delta.values()) <= 0:
                    raise exc.InputException(
                        'The property state, description, or env '
                        'is not provided for update.'
                    )

                # Description cannot be updated together with state.
                if delta.get('description') and delta.get('state'):
                    raise exc.InputException(
                        'The property description must be updated '
                        'separately from state.'
                    )

                # If state change, environment cannot be updated
                # if not RUNNING.
                if (delta.get('env') and
                        delta.get('state') and
                        delta['state'] != states.RUNNING):
                    raise exc.InputException(
                        'The property env can only be updated when workflow '
                        'execution is not running or on resume from pause.'
                    )

                if delta.get('description'):
                    wf_ex = db_api.update_workflow_execution(
                        id,
                        {'description': delta['description']}
                    )

                if not delta.get('state') and delta.get('env'):
                    wf_ex = db_api.get_workflow_execution(id)
                    wf_ex = wf_service.update_workflow_execution_env(
                        wf_ex,
                        delta.get('env')
                    )

                return delta, wf_ex

        delta, wf_ex = _compute_delta(wf_ex)

        if delta.get('state'):
            if states.is_paused(delta.get('state')):
                wf_ex = rpc.get_engine_client().pause_workflow(id)
            elif delta.get('state') == states.RUNNING:
                wf_ex = rpc.get_engine_client().resume_workflow(
                    id,
                    env=delta.get('env')
                )
            elif states.is_completed(delta.get('state')):
                msg = wf_ex.state_info if wf_ex.state_info else None
                wf_ex = rpc.get_engine_client().stop_workflow(
                    id,
                    delta.get('state'),
                    msg
                )
            else:
                # To prevent changing state in other cases throw a message.
                raise exc.InputException(
                    "Cannot change state to %s. Allowed states are: '%s" % (
                        wf_ex.state,
                        ', '.join([
                            states.RUNNING,
                            states.PAUSED,
                            states.SUCCESS,
                            states.ERROR,
                            states.CANCELLED
                        ])
                    )
                )

        return resources.Execution.from_dict(
            wf_ex if isinstance(wf_ex, dict) else wf_ex.to_dict()
        )

    @rest_utils.wrap_wsme_controller_exception
    @wsme_pecan.wsexpose(
        resources.Execution,
        body=resources.Execution,
        status_code=201
    )
    def post(self, wf_ex):
        """Create a new Execution.

        :param wf_ex: Execution object with input content.
        """
        acl.enforce('executions:create', context.ctx())

        LOG.info("Create execution [execution=%s]", wf_ex)

        exec_dict = wf_ex.to_dict()

        exec_id = exec_dict.get('id')

        if not exec_id:
            exec_id = uuidutils.generate_uuid()
            context.ctx(root_execution_id=exec_id)

            LOG.debug("Generated execution id [exec_id=%s]", exec_id)

            exec_dict.update({'id': exec_id})

            wf_ex = None
        else:
            # If ID is present we need to check if such execution exists.
            # If yes, the method just returns the object. If not, the ID
            # will be used to create a new execution.
            wf_ex = _get_workflow_execution(exec_id, must_exist=False)
            context.ctx(root_execution_id=exec_id)

            if wf_ex:
                return resources.Execution.from_db_model(wf_ex)

        source_execution_id = exec_dict.get('source_execution_id')

        source_exec_dict = None

        if source_execution_id:
            # If source execution is present we will perform a lookup for
            # previous workflow execution model and the information to start
            # a new workflow based on that information.
            source_exec_dict = db_api.get_workflow_execution(
                source_execution_id).to_dict()

            exec_dict['description'] = "{} Based on the execution '{}'".format(
                exec_dict['description'],
                source_execution_id
            )
            exec_dict['description'] = exec_dict['description'].strip()

        result_exec_dict = merge_dicts(source_exec_dict, exec_dict)

        if not (result_exec_dict.get('workflow_id') or
                result_exec_dict.get('workflow_name')):
            raise exc.WorkflowException(
                "Workflow ID or workflow name must be provided. Workflow ID is"
                " recommended."
            )

        engine = rpc.get_engine_client()

        result = engine.start_workflow(
            result_exec_dict.get(
                'workflow_id',
                result_exec_dict.get('workflow_name')
            ),
            result_exec_dict.get('workflow_namespace', ''),
            result_exec_dict.get('id'),
            result_exec_dict.get('input'),
            description=result_exec_dict.get('description', ''),
            **result_exec_dict.get('params') or {}
        )

        return resources.Execution.from_dict(result)

    @rest_utils.wrap_wsme_controller_exception
    @wsme_pecan.wsexpose(None, wtypes.text, bool, status_code=204)
    def delete(self, id, force=False):
        """Delete the specified Execution.

        :param id: UUID of execution to delete.
        :param force: Optional. Force the deletion of unfinished executions.
                      Default: false. While the api is backward compatible
                      the behaviour is not the same. The new default is the
                      safer option
        """
        acl.enforce('executions:delete', context.ctx())

        LOG.debug("Delete execution [id=%s]", id)

        if not force:
            state = db_api.get_workflow_execution(
                id,
                fields=(db_models.WorkflowExecution.state,)
            )[0]

            if not states.is_completed(state):
                raise exc.NotAllowedException(
                    "Only completed executions can be deleted. "
                    "Use --force to override this. "
                    "Execution {} is in {} state".format(id, state)
                )

        return rest_utils.rest_retry_on_db_error(
            db_api.delete_workflow_execution
        )(id)

    @rest_utils.wrap_wsme_controller_exception
    @wsme_pecan.wsexpose(resources.Executions, types.uuid, int,
                         types.uniquelist, types.list, types.uniquelist,
                         wtypes.text, types.uuid, wtypes.text,
                         types.uniquelist, types.jsontype, types.uuid,
                         types.uuid, STATE_TYPES, wtypes.text,
                         types.jsontype, types.jsontype, wtypes.text,
                         wtypes.text, bool, types.uuid,
                         bool, types.list)
    def get_all(self, marker=None, limit=None,
                sort_keys='created_at', sort_dirs='asc', fields='',
                workflow_name=None, workflow_id=None, description=None,
                tags=None, params=None, task_execution_id=None,
                root_execution_id=None, state=None, state_info=None,
                input=None, output=None, created_at=None,
                updated_at=None, include_output=None, project_id=None,
                all_projects=False, nulls=''):

        """Return all Executions.

        :param marker: Optional. Pagination marker for large data sets.
        :param limit: Optional. Maximum number of resources to return in a
                      single result. Default value is None for backward
                      compatibility.
        :param sort_keys: Optional. Columns to sort results by.
                          Default: created_at, which is backward compatible.
        :param sort_dirs: Optional. Directions to sort corresponding to
                          sort_keys, "asc" or "desc" can be chosen.
                          Default: desc. The length of sort_dirs can be equal
                          or less than that of sort_keys.
        :param fields: Optional. A specified list of fields of the resource to
                       be returned. 'id' will be included automatically in
                       fields if it's not provided, since it will be used when
                       constructing 'next' link.
        :param workflow_name: Optional. Keep only resources with a specific
                              workflow name.
        :param workflow_id: Optional. Keep only resources with a specific
                            workflow ID.
        :param description: Optional. Keep only resources with a specific
                            description.
        :param tags: Optional. Keep only resources containing specific tags.
        :param params: Optional. Keep only resources with specific parameters.
        :param task_execution_id: Optional. Keep only resources with a
                                  specific task execution ID.
        :param root_execution_id: Optional. Keep only resources with a
                                  specific root execution ID.
        :param state: Optional. Keep only resources with a specific state.
        :param state_info: Optional. Keep only resources with specific
                           state information.
        :param input: Optional. Keep only resources with a specific input.
        :param output: Optional. Keep only resources with a specific output.
        :param created_at: Optional. Keep only resources created at a specific
                           time and date.
        :param updated_at: Optional. Keep only resources with specific latest
                           update time and date.
        :param include_output: Optional. Include the output for all executions
                               in the list.
        :param project_id: Optional. Only get executions belong to the project.
            Admin required.
        :param all_projects: Optional. Get resources of all projects. Admin
            required.
        :param nulls: Optional. The names of the columns with null value in
                        the query.
        """
        acl.enforce('executions:list', context.ctx())

        db_models.WorkflowExecution.check_allowed_none_values(nulls)

        if all_projects or project_id:
            acl.enforce('executions:list:all_projects', context.ctx())

        filters = filter_utils.create_filters_from_request_params(
            none_values=nulls,
            created_at=created_at,
            workflow_name=workflow_name,
            workflow_id=workflow_id,
            tags=tags,
            params=params,
            task_execution_id=task_execution_id,
            state=state,
            state_info=state_info,
            input=input,
            output=output,
            updated_at=updated_at,
            description=description,
            project_id=project_id,
            root_execution_id=root_execution_id,
        )

        LOG.debug(
            "Fetch executions. marker=%s, limit=%s, sort_keys=%s, "
            "sort_dirs=%s, filters=%s, all_projects=%s", marker, limit,
            sort_keys, sort_dirs, filters, all_projects
        )

        if include_output:
            resource_function = _get_workflow_execution_resource_with_output
        else:
            resource_function = _get_workflow_execution_resource

        return rest_utils.get_all(
            resources.Executions,
            resources.Execution,
            db_api.get_workflow_executions,
            db_api.get_workflow_execution,
            resource_function=resource_function,
            marker=marker,
            limit=limit,
            sort_keys=sort_keys,
            sort_dirs=sort_dirs,
            fields=fields,
            all_projects=all_projects,
            **filters
        )



# Copyright 2020 - Nokia Software.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from mistral.db.v2 import api as db_api
from mistral.services import actions
from mistral.tests.unit.api import base


TEST_MODULE_TEXT = """
from mistral_lib import actions

class DummyAction(actions.Action):
    def run(self, context):
        return "Hello from the dummy action 1!"

    def test(self, context):
        return None

class DummyAction2(actions.Action):
    def run(self, context):
        return "Hello from the dummy action 2!"

    def test(self, context):
        return None"""


class TestDynamicActionsController(base.APITest):
    def setUp(self):
        super(TestDynamicActionsController, self).setUp()

        resp = self.app.post(
            '/v2/code_sources?name=test_dummy_module',
            TEST_MODULE_TEXT
        )

        self.code_source_id = resp.json['id']

        self.addCleanup(db_api.delete_code_sources)
        self.addCleanup(db_api.delete_dynamic_action_definitions)

    def _create_dynamic_action(self):
        return self.app.post_json(
            '/v2/dynamic_actions',
            {
                'name': 'dummy_action',
                'class_name': 'DummyAction',
                'code_source_id': self.code_source_id
            }
        )

    def test_post(self):
        resp = self.app.post_json(
            '/v2/dynamic_actions',
            {
                'name': 'dummy_action',
                'class_name': 'DummyAction',
                'code_source_id': self.code_source_id
            }
        )

        # Check the structure of the response.
        self.assertEqual(201, resp.status_int)

        dyn_action = resp.json

        self.assertEqual('dummy_action', dyn_action['name'])
        self.assertEqual('DummyAction', dyn_action['class_name'])
        self.assertEqual(self.code_source_id, dyn_action['code_source_id'])

        # Make sure the action can be found via the system action provider
        # and it's fully functioning.
        provider = actions.get_system_action_provider()

        action_desc = provider.find('dummy_action')

        self.assertIsNotNone(action_desc)

        action = action_desc.instantiate({}, None)

        self.assertIsNotNone(action)

        self.assertEqual("Hello from the dummy action 1!", action.run(None))

    def test_put(self):
        resp = self._create_dynamic_action()

        self.assertEqual(201, resp.status_int)

        resp = self.app.put_json(
            '/v2/dynamic_actions',
            {
                'name': 'dummy_action',
                'class_name': 'NewDummyAction'
            }
        )

        self.assertEqual(200, resp.status_int)

        dyn_action = resp.json

        self.assertEqual('dummy_action', dyn_action['name'])
        self.assertEqual('NewDummyAction', dyn_action['class_name'])

    def test_get(self):
        self._create_dynamic_action()

        resp = self.app.get('/v2/dynamic_actions/dummy_action')

        self.assertEqual(200, resp.status_int)

        dyn_action = resp.json

        self.assertEqual('dummy_action', dyn_action['name'])
        self.assertEqual('DummyAction', dyn_action['class_name'])
        self.assertEqual(self.code_source_id, dyn_action['code_source_id'])
        self.assertEqual('test_dummy_module', dyn_action['code_source_name'])

    def test_get_all(self):
        self._create_dynamic_action()

        resp = self.app.get('/v2/dynamic_actions')

        resp_json = resp.json

        self.assertEqual(200, resp.status_int)

        dynamic_actions = resp_json.get('dynamic_actions')

        self.assertEqual(1, len(dynamic_actions))

    def test_delete(self):
        resp = self._create_dynamic_action()

        # Check the structure of the response
        self.assertEqual(201, resp.status_int)

        resp = self.app.get('/v2/dynamic_actions/dummy_action')

        self.assertEqual(200, resp.status_int)

        self.app.delete('/v2/dynamic_actions/dummy_action')

        resp = self.app.get(
            '/v2/dynamic_actions/dummy_action',
            expect_errors=True
        )

        self.assertEqual(404, resp.status_int)

        # Make sure the system action provider doesn't find an action
        # descriptor for the action.
        provider = actions.get_system_action_provider()

        self.assertIsNone(provider.find('dummy_action'))



# Copyright 2020 Nokia Software.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from mistral.actions import dynamic_action
from mistral.db.v2 import api as db_api
from mistral.tests.unit import base

DUMMY_CODE_SOURCE = """from mistral_lib import actions

class DummyAction(actions.Action):
    def run(self, context):
        return None
    def test(self, context):
        return None
class DummyAction2(actions.Action):
    def run(self, context):
        return None
    def test(self, context):
        return None
"""

ACTIONS = []
NAMESPACE = "ns"


class DynamicActionProviderTest(base.DbTestCase):
    def _create_code_source(self, namespace=''):
        return db_api.create_code_source(
            {
                'name': 'code_source',
                'content': DUMMY_CODE_SOURCE,
                'namespace': namespace,
                'version': 0
            }
        )

    def _delete_code_source(self):
        return db_api.delete_code_source('code_source')

    def _create_dynamic_actions(self, code_source, namespace=''):
        db_api.create_dynamic_action_definition(
            {
                "name": "dummy_action",
                "namespace": namespace,
                "class_name": "DummyAction",
                "code_source_id": code_source.id,
                "code_source_name": code_source.name
            }
        )
        db_api.create_dynamic_action_definition(
            {
                "name": "dummy_action2",
                "namespace": namespace,
                "class_name": "DummyAction2",
                "code_source_id": code_source.id,
                "code_source_name": code_source.name
            }
        )

    def test_dynamic_actions(self):
        provider = dynamic_action.DynamicActionProvider()

        action_descs = provider.find_all()

        self.assertEqual(0, len(action_descs))

        code_source = self._create_code_source()

        self._create_dynamic_actions(code_source)

        action_descs = provider.find_all()

        self.assertEqual(2, len(action_descs))

        self._delete_code_source()

    def test_loaded_actions_deleted_from_db(self):
        provider = dynamic_action.DynamicActionProvider()

        action_descs = provider.find_all()

        self.assertEqual(0, len(action_descs))

        code_source = self._create_code_source()

        self._create_dynamic_actions(code_source)

        action_descs = provider.find_all()

        self.assertEqual(2, len(action_descs))

        self._delete_code_source()

        action_descs = provider.find_all()

        self.assertEqual(0, len(action_descs))

    def test_dynamic_actions_with_namespace(self):
        provider = dynamic_action.DynamicActionProvider()

        action_descs = provider.find_all()

        self.assertEqual(0, len(action_descs))

        code_source = self._create_code_source()

        self._create_dynamic_actions(
            code_source=code_source,
            namespace=NAMESPACE
        )

        action_descs = provider.find_all(namespace=NAMESPACE)

        self.assertEqual(2, len(action_descs))

        action_descs = provider.find_all(namespace='')

        self.assertEqual(0, len(action_descs))

        self._delete_code_source()



# Copyright 2014 - Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import sys

import keystonemiddleware.opts as keystonemw_opts
from oslo_config import cfg
from oslo_log import log as logging

from mistral import config
from mistral.db.v2 import api as db_api
from mistral.services import action_manager
from mistral.services import workflows

CONF = cfg.CONF
LOG = logging.getLogger(__name__)


def main():
    # NOTE(jaosorior): This is needed in order for db-sync to also register the
    # keystonemiddleware options. Those options are used by clients that need a
    # keystone session in order to be able to register their actions.
    # This can be removed when mistral moves out of using keystonemiddleware in
    # favor of keystoneauth1.
    for group, opts in keystonemw_opts.list_auth_token_opts():
        CONF.register_opts(opts, group=group)

    logging.register_options(CONF)

    config.parse_args()

    if len(CONF.config_file) == 0:
        print("Usage: sync_db --config-file <path-to-config-file>")
        return exit(1)
    logging.setup(CONF, 'Mistral')

    LOG.info("Starting db_sync")

    LOG.debug("Setting up db")
    db_api.setup_db()

    LOG.debug("populating db")
    action_manager.sync_db()
    workflows.sync_db()


if __name__ == '__main__':
    sys.exit(main())



# Copyright 2016 OpenStack Foundation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""add type to task execution

Revision ID: 020
Revises: 019
Create Date: 2016-10-05 13:24:52.911011

"""

# revision identifiers, used by Alembic.
revision = '020'
down_revision = '019'

from alembic import op
from mistral.db.sqlalchemy import types as st
import sqlalchemy as sa

# A simple model of the task executions table with only the fields needed for
# the migration.
task_executions = sa.Table(
    'task_executions_v2',
    sa.MetaData(),
    sa.Column('id', sa.String(36), nullable=False),
    sa.Column(
        'spec',
        st.JsonMediumDictType()
    ),
    sa.Column('type', sa.String(10), nullable=True)
)


def upgrade():

    op.add_column(
        'task_executions_v2',
        sa.Column('type', sa.String(length=10), nullable=True)
    )

    session = sa.orm.Session(bind=op.get_bind())
    values = []

    for row in session.query(task_executions):
        values.append({'id': row[0],
                       'spec': row[1]})

    with session.begin(subtransactions=True):
        for value in values:
            task_type = "ACTION"
            if "workflow" in value['spec']:
                task_type = "WORKFLOW"

            session.execute(
                task_executions.update().values(type=task_type).where(
                    task_executions.c.id == value['id']
                )
            )

    # this commit appears to be necessary
    session.commit()



# Copyright 2020 Nokia Software.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from mistral.services import actions
from mistral.tests.unit import base


class LegacyActionProviderTest(base.DbTestCase):
    def test_get_system_action_provider(self):
        self.override_config(
            'load_action_generators',
            False,
            'legacy_action_provider'
        )
        self.override_config(
            'only_builtin_actions',
            True,
            'legacy_action_provider'
        )

        system_provider = actions.get_system_action_provider()

        action_descs = system_provider.find_all()

        for a in action_descs:
            print(a)

        self.assertTrue(len(action_descs) > 0)

        self.assertTrue(
            all(
                [
                    a_d.action_class.__module__.startswith('mistral.')
                    for a_d in action_descs
                ]
            )
        )



# Copyright 2020 Nokia Software.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from mistral.actions import adhoc
from mistral.services import adhoc_actions as adhoc_action_service
from mistral.tests.unit import base


class AdHocActionProviderTest(base.DbTestCase):
    def test_adhoc_actions(self):
        provider = adhoc.AdHocActionProvider()

        action_descs = provider.find_all()

        self.assertEqual(0, len(action_descs))

        action_txt = """
        version: '2.0'

        my_adhoc_action:
          base: std.echo
          base-input:
            output: "<% $.s1 %>+<% $.s2 %>"
          input:
            - s1: "a"
            - s2
          output: "<% $ %> and <% $ %>"
        """

        adhoc_action_service.create_actions(action_txt)

        action_descs = provider.find_all()

        self.assertEqual(1, len(action_descs))

        action_desc = action_descs[0]

        self.assertEqual('my_adhoc_action', action_desc.name)
        self.assertEqual(action_txt, action_desc.definition)



# Copyright 2018 - Nokia Networks.
#
# Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.


from oslo_config import cfg
from oslo_log import log as logging

from mistral.rpc import base as rpc
from mistral.service import base as service_base


LOG = logging.getLogger(__name__)

CONF = cfg.CONF


class SchedulerServer(service_base.MistralService):
    """Scheduler server.

    Manages scheduler life-cycle and gets registered as an RPC
    endpoint to process scheduler specific calls.
    """

    def __init__(self, scheduler, setup_profiler=True):
        super(SchedulerServer, self).__init__(
            'scheduler_group',
            setup_profiler
        )

        self.scheduler = scheduler
        self._rpc_server = None

    def start(self):
        super(SchedulerServer, self).start()

        self._rpc_server = rpc.get_rpc_server_driver()(cfg.CONF.engine)
        self._rpc_server.register_endpoint(self)

        self._rpc_server.run()

        self._notify_started('Scheduler server started.')

    def stop(self, graceful=False):
        super(SchedulerServer, self).stop()

        if self._rpc_server:
            self._rpc_server.stop(graceful)

    def schedule(self, rpc_ctx, job):
        """Receives requests over RPC to schedule delayed calls.

        :param rpc_ctx: RPC request context.
        :param job: Scheduler job.
        """
        LOG.info("Received RPC request 'schedule'[job=%s]", job)

        return self.scheduler.schedule(job, allow_redistribute=False)



# Copyright 2016 NEC Corporation. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import fixtures

from mistral.api import access_control as acl
from mistral import policies
from oslo_config import cfg
from oslo_policy import opts as policy_opts
from oslo_policy import policy as oslo_policy


class PolicyFixture(fixtures.Fixture):
    def setUp(self):
        super(PolicyFixture, self).setUp()

        cfg.CONF(args=[], project='mistral')

        policy_opts.set_defaults(cfg.CONF)

        acl._ENFORCER = oslo_policy.Enforcer(cfg.CONF)
        acl._ENFORCER.register_defaults(policies.list_rules())
        acl._ENFORCER.load_rules()

        self.addCleanup(acl._ENFORCER.clear)

    def register_rules(self, rules):
        enf = acl._ENFORCER

        for rule_name, rule_check_str in rules.items():
            enf.register_default(oslo_policy.RuleDefault(rule_name,
                                                         rule_check_str))

    def change_policy_definition(self, rules):
        enf = acl._ENFORCER

        for rule_name, rule_check_str in rules.items():
            enf.rules[rule_name] = oslo_policy.RuleDefault(
                rule_name, rule_check_str).check



# Copyright 2016 - Nokia Networks.
# Copyright 2016 - Brocade Communications Systems, Inc.
# Copyright 2018 - Extreme Networks, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import abc
from oslo_config import cfg
from oslo_log import log as logging
from osprofiler import profiler

from mistral.db.v2 import api as db_api
from mistral.engine import post_tx_queue
from mistral.engine import utils as engine_utils
from mistral.engine import workflow_handler as wf_handler
from mistral import exceptions as exc
from mistral.executors import base as exe
from mistral.lang import parser as spec_parser
from mistral.rpc import clients as rpc
from mistral.services import security
from mistral.utils import wf_trace
from mistral.workflow import data_flow
from mistral.workflow import states
from mistral_lib import utils


LOG = logging.getLogger(__name__)
CONF = cfg.CONF


class Action(object, metaclass=abc.ABCMeta):
    """Action.

    Represents a workflow action and defines interface that can be used by
    Mistral engine or its components in order to manipulate with actions.
    """

    def __init__(self, action_desc, action_ex=None, task_ex=None,
                 task_ctx=None):
        self.action_desc = action_desc
        self.action_ex = action_ex
        self.namespace = action_desc.namespace if action_desc else None
        self.task_ex = action_ex.task_execution if action_ex else task_ex
        self.task_ctx = task_ctx

    @abc.abstractmethod
    def complete(self, result):
        """Complete action and process its result.

        :param result: Action result.
        """
        raise NotImplementedError

    def fail(self, msg):
        assert self.action_ex

        # When we set an ERROR state we should safely set output value getting
        # w/o exceptions due to field size limitations.
        msg = utils.cut_by_kb(
            msg,
            cfg.CONF.engine.execution_field_size_limit_kb
        )

        self.action_ex.state = states.ERROR
        self.action_ex.output = {'result': msg}

    def update(self, state):
        assert self.action_ex

        # TODO(rakhmerov): Not sure we can do it for all actions.
        action = self.action_desc.instantiate(self.action_ex.input, {})

        if state == states.PAUSED and action.is_sync():
            raise exc.InvalidStateTransitionException(
                'Transition to the PAUSED state is only supported '
                'for asynchronous action execution.'
            )

        if not states.is_valid_transition(self.action_ex.state, state):
            raise exc.InvalidStateTransitionException(
                'Invalid state transition from %s to %s.' %
                (self.action_ex.state, state)
            )

        self.action_ex.state = state

    @abc.abstractmethod
    def schedule(self, input_dict, target, index=0, desc='', safe_rerun=False,
                 timeout=None):
        """Schedule action run.

        This method is needed to schedule action run so its result can
        be received later by engine. In this sense it will be running in
        asynchronous mode from engine perspective (don't confuse with
        executor asynchrony when executor doesn't immediately send a
        result).

        :param timeout: a period of time in seconds after which execution of
            action will be interrupted
        :param input_dict: Action input.
        :param target: Target (group of action executors).
        :param index: Action execution index. Makes sense for some types.
        :param desc: Action execution description.
        :param safe_rerun: If true, action would be re-run if executor dies
            during execution.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def run(self, input_dict, target, index=0, desc='', save=True,
            safe_rerun=False, timeout=None):
        """Immediately run action.

        This method runs method w/o scheduling its run for a later time.
        From engine perspective action will be processed in synchronous
        mode.

        :param timeout: a period of time in seconds after which execution of
            action will be interrupted
        :param input_dict: Action input.
        :param target: Target (group of action executors).
        :param index: Action execution index. Makes sense for some types.
        :param desc: Action execution description.
        :param save: True if action execution object needs to be saved.
        :param safe_rerun: If true, action would be re-run if executor dies
            during execution.
        :return: Action output.
        """
        raise NotImplementedError

    def _prepare_execution_context(self):
        res = {}

        if self.task_ex:
            wf_ex = self.task_ex.workflow_execution

            res['workflow_execution_id'] = wf_ex.id
            res['task_execution_id'] = self.task_ex.id
            res['workflow_name'] = wf_ex.name

        if self.action_ex:
            res['action_execution_id'] = self.action_ex.id
            res['callback_url'] = (
                '/v2/action_executions/%s' % self.action_ex.id
            )

        return res

    def _create_action_execution(self, input_dict, runtime_ctx,
                                 desc='', action_ex_id=None, is_sync=True):
        action_ex_id = action_ex_id or utils.generate_unicode_uuid()

        values = {
            'id': action_ex_id,
            'name': self.action_desc.name,
            'state': states.RUNNING,
            'input': input_dict,
            'runtime_context': runtime_ctx,
            'workflow_namespace': self.namespace,
            'description': desc,
            'is_sync': is_sync
        }

        if self.task_ex:
            values.update({
                'task_execution_id': self.task_ex.id,
                'workflow_name': self.task_ex.workflow_name,
                'workflow_namespace': self.task_ex.workflow_namespace,
                'workflow_id': self.task_ex.workflow_id,
                'project_id': self.task_ex.project_id,
            })
        else:
            values.update({
                'project_id': security.get_project_id(),
            })

        LOG.info("Create action execution [action_name=%s, action_ex_id=%s]",
                 self.action_desc.name, action_ex_id)

        self.action_ex = db_api.create_action_execution(values)

        if self.task_ex:
            # Add to collection explicitly so that it's in a proper
            # state within the current session.
            self.task_ex.action_executions.append(self.action_ex)

    @profiler.trace('action-log-result', hide_args=True)
    def _log_result(self, prev_state, result):
        state = self.action_ex.state

        if prev_state != state:
            wf_trace.info(
                None,
                "Action '%s' (%s)(task_name=%s, "
                "task_ex_id=%s) [%s -> %s, %s]" %
                (self.action_ex.name,
                 self.action_ex.id,
                 self.task_ex.name if self.task_ex else None,
                 self.task_ex.id if self.task_ex else None,
                 prev_state,
                 state,
                 result.cut_repr())
            )


class RegularAction(Action):
    """Regular Python action."""

    @profiler.trace('regular-action-complete', hide_args=True)
    def complete(self, result):
        assert self.action_ex

        if states.is_completed(self.action_ex.state):
            raise ValueError(
                "Action {} is already completed".format(self.action_ex.id)
            )

        prev_state = self.action_ex.state

        if result.is_success():
            self.action_ex.state = states.SUCCESS
        elif result.is_cancel():
            self.action_ex.state = states.CANCELLED
        else:
            self.action_ex.state = states.ERROR

        # Convert the result, if needed.
        converted_result = self.action_desc.post_process_result(result)

        self.action_ex.output = converted_result.to_dict()
        self.action_ex.accepted = True

        self._log_result(prev_state, result)

    @profiler.trace('action-schedule', hide_args=True)
    def schedule(self, input_dict, target, index=0, desc='', safe_rerun=False,
                 timeout=None):
        assert not self.action_ex

        self.action_desc.check_parameters(input_dict)

        wf_ex = self.task_ex.workflow_execution if self.task_ex else None

        wf_ctx = data_flow.ContextView(
            self.task_ctx,
            data_flow.get_workflow_environment_dict(wf_ex),
            wf_ex.context if wf_ex else {}
        )

        try:
            action = self.action_desc.instantiate(input_dict, wf_ctx)
        except Exception:
            raise exc.InvalidActionException(
                'Failed to instantiate an action'
                ' [action_desc=%s, input_dict=%s]'
                % (self.action_desc, input_dict)
            )

        # Assign the action execution ID here to minimize database calls.
        # Otherwise, the input property of the action execution DB object needs
        # to be updated with the action execution ID after the action execution
        # DB object is created.
        action_ex_id = utils.generate_unicode_uuid()

        self._create_action_execution(
            input_dict,
            self._prepare_runtime_context(index, safe_rerun),
            desc=desc,
            action_ex_id=action_ex_id,
            is_sync=action.is_sync()
        )

        def _run_action():
            executor = exe.get_executor(cfg.CONF.executor.type)

            return executor.run_action(
                action,
                self.action_ex.id if self.action_ex is not None else None,
                safe_rerun,
                self._prepare_execution_context(),
                target=target,
                timeout=timeout
            )

        # Register an asynchronous command to run the action
        # on an executor outside of the main DB transaction.
        post_tx_queue.register_operation(_run_action)

    @profiler.trace('action-run', hide_args=True)
    def run(self, input_dict, target, index=0, desc='', save=True,
            safe_rerun=False, timeout=None):
        assert not self.action_ex

        self.action_desc.check_parameters(input_dict)

        try:
            action = self.action_desc.instantiate(input_dict, {})
        except Exception:
            raise exc.InvalidActionException(
                'Failed to instantiate an action'
                ' [action_desc=%s, input_dict=%s]'
                % (self.action_desc, input_dict)
            )

        # Assign the action execution ID here to minimize database calls.
        # Otherwise, the input property of the action execution DB object needs
        # to be updated with the action execution ID after the action execution
        # DB object is created.
        action_ex_id = utils.generate_unicode_uuid()

        if save:
            self._create_action_execution(
                input_dict,
                self._prepare_runtime_context(index, safe_rerun),
                desc=desc,
                action_ex_id=action_ex_id,
                is_sync=action.is_sync()
            )

        executor = exe.get_executor(cfg.CONF.executor.type)

        return executor.run_action(
            action,
            self.action_ex.id if self.action_ex is not None else None,
            safe_rerun,
            self._prepare_execution_context(),
            target=target,
            async_=False,
            timeout=timeout
        )

    def _prepare_runtime_context(self, index, safe_rerun):
        """Template method to prepare action runtime context.

        Regular action inserts an index into its runtime context and
        the flag showing if the action is safe to rerun (i.e. it's
        idempotent).
        """
        return {'index': index, 'safe_rerun': safe_rerun}


class WorkflowAction(Action):
    """Workflow action."""

    def __init__(self, wf_name, **kwargs):
        super(WorkflowAction, self).__init__(None, **kwargs)

        self.wf_name = wf_name

    @profiler.trace('workflow-action-complete', hide_args=True)
    def complete(self, result):
        # No-op because in case of workflow result is already processed.
        pass

    @profiler.trace('workflow-action-schedule', hide_args=True)
    def schedule(self, input_dict, target, index=0, desc='', safe_rerun=False,
                 timeout=None):
        assert not self.action_ex

        self.validate_input(input_dict)

        parent_wf_ex = self.task_ex.workflow_execution
        parent_wf_spec = spec_parser.get_workflow_spec_by_execution_id(
            parent_wf_ex.id
        )

        wf_def = engine_utils.resolve_workflow_definition(
            parent_wf_ex.workflow_name,
            parent_wf_spec.get_name(),
            namespace=parent_wf_ex.params['namespace'],
            wf_spec_name=self.wf_name
        )

        wf_spec = spec_parser.get_workflow_spec_by_definition_id(
            wf_def.id,
            wf_def.updated_at
        )

        # If the parent has a root_execution_id, it must be a sub-workflow. So
        # we should propagate that ID down. Otherwise the parent must be the
        # root execution and we should use the parents ID.
        root_execution_id = parent_wf_ex.root_execution_id or parent_wf_ex.id

        wf_params = {
            'root_execution_id': root_execution_id,
            'task_execution_id': self.task_ex.id,
            'index': index,
            'namespace': parent_wf_ex.params['namespace']
        }

        if 'notify' in parent_wf_ex.params:
            wf_params['notify'] = parent_wf_ex.params['notify']

        for k, v in list(input_dict.items()):
            if k not in wf_spec.get_input():
                wf_params[k] = v
                del input_dict[k]

        if cfg.CONF.engine.start_subworkflows_via_rpc:
            def _start_subworkflow():
                rpc.get_engine_client().start_workflow(
                    wf_def.id,
                    wf_def.namespace,
                    None,
                    input_dict,
                    "sub-workflow execution",
                    async_=True,
                    **wf_params
                )

            post_tx_queue.register_operation(_start_subworkflow)
        else:
            wf_handler.start_workflow(
                wf_def.id,
                wf_def.namespace,
                None,
                input_dict,
                "sub-workflow execution",
                wf_params
            )

    @profiler.trace('workflow-action-run', hide_args=True)
    def run(self, input_dict, target, index=0, desc='', save=True,
            safe_rerun=True, timeout=None):
        raise NotImplementedError('Does not apply to this WorkflowAction.')

    def validate_input(self, input_dict):
        # TODO(rakhmerov): Implement.
        pass



# Copyright 2014 - Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from oslo_config import cfg

from mistral.db.v2 import api as db_api
from mistral import exceptions as exc
from mistral.services import workbooks as wb_service
from mistral.services import workflows as wf_service
from mistral.tests.unit.engine import base
from mistral.workflow import states


# Use the set_default method to set value otherwise in certain test cases
# the change in value is not permanent.
cfg.CONF.set_default('auth_enable', False, group='pecan')


WORKBOOK = """
---
version: '2.0'

name: my_wb

workflows:
  wf1:
    type: reverse
    input:
      - param1
      - param2

    tasks:
      task1:
        action: std.echo output=<% $.param1 %>
        publish:
          result1: <% task(task1).result %>

      task2:
        action: std.echo output="<% $.result1 %> & <% $.param2 %>"
        publish:
          result2: <% task(task2).result %>
        requires: [task1]

      task3:
        action: std.noop

      task4:
        action: std.noop
        requires: task3
"""


class ReverseWorkflowEngineTest(base.EngineTestCase):
    def setUp(self):
        super(ReverseWorkflowEngineTest, self).setUp()

        wb_service.create_workbook_v2(WORKBOOK)

    def test_start_task1(self):
        wf_input = {
            'param1': 'a',
            'param2': 'b'
        }

        wf_ex = self.engine.start_workflow(
            'my_wb.wf1',
            wf_input=wf_input,
            task_name='task1'
        )

        # Execution 1.
        self.assertIsNotNone(wf_ex)
        self.assertDictEqual(wf_input, wf_ex.input)
        self.assertDictEqual(
            {
                'task_name': 'task1',
                'namespace': '',
                'env': {}
            },
            wf_ex.params
        )

        # Wait till workflow 'wf1' is completed.
        self.await_workflow_success(wf_ex.id)

        with db_api.transaction():
            wf_ex = db_api.get_workflow_execution(wf_ex.id)

            task_execs = wf_ex.task_executions

        self.assertEqual(1, len(task_execs))
        self.assertEqual(1, len(db_api.get_task_executions()))

        task_ex = self._assert_single_item(
            task_execs,
            name='task1',
            state=states.SUCCESS
        )

        self.assertDictEqual({'result1': 'a'}, task_ex.published)

    def test_start_task2(self):
        wf_input = {
            'param1': 'a',
            'param2': 'b'
        }

        wf_ex = self.engine.start_workflow(
            'my_wb.wf1',
            wf_input=wf_input,
            task_name='task2'
        )

        # Execution 1.
        self.assertIsNotNone(wf_ex)
        self.assertDictEqual(wf_input, wf_ex.input)
        self.assertDictEqual(
            {
                'task_name': 'task2',
                'namespace': '',
                'env': {}
            },
            wf_ex.params
        )

        # Wait till workflow 'wf1' is completed.
        self.await_workflow_success(wf_ex.id)

        with db_api.transaction():
            wf_ex = db_api.get_workflow_execution(wf_ex.id)

            task_execs = wf_ex.task_executions

        self.assertEqual(2, len(task_execs))
        self.assertEqual(2, len(db_api.get_task_executions()))

        task1_ex = self._assert_single_item(
            task_execs,
            name='task1',
            state=states.SUCCESS
        )

        self.assertDictEqual({'result1': 'a'}, task1_ex.published)

        task2_ex = self._assert_single_item(
            task_execs,
            name='task2',
            state=states.SUCCESS
        )

        self.assertDictEqual({'result2': 'a & b'}, task2_ex.published)

    def test_one_line_requires_syntax(self):
        wf_input = {'param1': 'a', 'param2': 'b'}

        wf_ex = self.engine.start_workflow(
            'my_wb.wf1',
            wf_input=wf_input,
            task_name='task4'
        )

        self.await_workflow_success(wf_ex.id)

        tasks = db_api.get_task_executions()

        self.assertEqual(2, len(tasks))

        self._assert_single_item(tasks, name='task4', state=states.SUCCESS)
        self._assert_single_item(tasks, name='task3', state=states.SUCCESS)

    def test_inconsistent_task_names(self):
        wf_text = """
        version: '2.0'

        wf:
          type: reverse

          tasks:
            task2:
              action: std.noop

            task3:
              action: std.noop
              requires: [task1]
        """

        exception = self.assertRaises(
            exc.InvalidModelException,
            wf_service.create_workflows,
            wf_text
        )

        self.assertIn("Task 'task1' not found", str(exception))



# Copyright 2014 - Mirantis, Inc.
# Copyright 2015 - StackStorm, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from mistral.lang import types
from mistral.lang.v2 import base
from mistral.lang.v2 import publish

NEXT_TASK = {
    "oneOf": [
        {
            "type": "string",
            "pattern": r"^\S+$",
            "description": "Task name (e.g.: `task1`)"
        },
        {
            "type": "string",
            "pattern": r"^\w+ \w+=(.*)$",
            "description": "Task name with dict parameter "
                           "(e.g.: `fail msg=\"test\"`, "
                           "`fail msg=<% task() %>`)"
        },
        {
            "type": "string",
            "pattern": r"^\w+\(\w+=(.*)\)$",
            "description": "Task name with func parameter "
                           "(e.g.: `fail(msg=\"test\")`, "
                           "`fail(msg=<% task() %>)`)"
        }
    ]
}

TASK_WITH_EXPRESSION = {
    "type": "object",
    "minProperties": 1,
    "maxProperties": 1,
    "patternProperties": {
        x['pattern']: types.EXPRESSION for x in NEXT_TASK['oneOf']
    },
    "description": "All next task variants plus expression (e.g.: "
                   "`task1: <% $.vm_id != null %>`, "
                   "`fail(msg=\"test\"): <% $.vm_id != null %>`)."
}

LIST_OF_TASKS = {
    "type": "array",
    "items": {
        "oneOf": [
            NEXT_TASK,
            TASK_WITH_EXPRESSION
        ]
    },
    "uniqueItems": True,
    "minItems": 1
}

ADVANCED_PUBLISHING_DICT = {
    "type": "object",
    "minProperties": 1,
    "properties": {
        "publish": publish.PublishSpec.get_schema(),
        "next": {
            "oneOf": [
                NEXT_TASK,
                TASK_WITH_EXPRESSION,
                LIST_OF_TASKS
            ]
        }
    },
    "additionalProperties": False
}


class OnClauseSpec(base.BaseSpec):
    _schema = {
        "oneOf": [
            NEXT_TASK,
            TASK_WITH_EXPRESSION,
            LIST_OF_TASKS,
            ADVANCED_PUBLISHING_DICT
        ]
    }

    def __init__(self, data, validate):
        super(OnClauseSpec, self).__init__(data, validate)

        if not isinstance(data, dict):
            # Old simple schema.
            self._publish = None
            self._next = prepare_next_clause(data)
        else:
            # New advanced schema.
            self._publish = self._spec_property('publish', publish.PublishSpec)
            self._next = prepare_next_clause(data.get('next'))

    @classmethod
    def get_schema(cls, includes=('definitions',)):
        return super(OnClauseSpec, cls).get_schema(includes)

    def get_publish(self):
        return self._publish

    def get_next(self):
        return self._next


def _as_list_of_tuples(data):
    if not data:
        return []

    if isinstance(data, str):
        return [_as_tuple(data)]

    return [_as_tuple(item) for item in data]


def _as_tuple(val):
    return list(val.items())[0] if isinstance(val, dict) else (val, '')


def prepare_next_clause(next_clause):
    list_of_tuples = _as_list_of_tuples(next_clause)

    for i, task in enumerate(list_of_tuples):
        task_name, params = OnClauseSpec._parse_cmd_and_input(task[0])

        list_of_tuples[i] = (task_name, task[1], params)

    return list_of_tuples



# Copyright 2015 - Mirantis, Inc.
# Copyright 2016 - Brocade Communications Systems, Inc.
# Copyright 2020 Nokia Software.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from oslo_config import cfg
from oslo_log import log as logging
from pecan import rest
from wsme import types as wtypes
import wsmeext.pecan as wsme_pecan

from mistral.api import access_control as acl
from mistral.api.controllers.v2 import resources
from mistral.api.controllers.v2 import types
from mistral import context
from mistral.db.v2 import api as db_api
from mistral import exceptions as exc
from mistral.rpc import clients as rpc
from mistral.utils import filter_utils
from mistral.utils import rest_utils
from mistral.workflow import states
from mistral_lib import actions as ml_actions


LOG = logging.getLogger(__name__)

SUPPORTED_TRANSITION_STATES = [
    states.SUCCESS,
    states.ERROR,
    states.CANCELLED,
    states.PAUSED,
    states.RUNNING
]


def _load_deferred_output_field(action_ex):
    # We need to refer to this lazy-load field explicitly in
    # order to make sure that it is correctly loaded.
    hasattr(action_ex, 'output')


# Use retries to prevent possible failures.
@rest_utils.rest_retry_on_db_error
def _get_action_execution(id, fields=()):
    if fields and 'id' not in fields:
        fields.insert(0, 'id')

    with db_api.transaction():
        return _get_action_execution_resource(
            db_api.get_action_execution(id),
            fields=fields
        )


def _get_action_execution_resource(action_ex, fields=()):
    _load_deferred_output_field(action_ex)

    if fields and 'id' not in fields:
        fields.insert(0, 'id')

    return _get_action_execution_resource_for_list(action_ex, fields=fields)


def _get_action_execution_resource_for_list(action_ex, fields=()):
    # TODO(nmakhotkin): Get rid of using dicts for constructing resources.
    # TODO(nmakhotkin): Use db_model for this instead.

    # field_task_name_needed = 'task_name' in fields
    # if field_task_name_needed:
    #     fields.remove('task_name')

    res = resources.ActionExecution.from_db_model(action_ex, fields=fields)

    # if not fields or field_task_name_needed:
    #     task_name = (action_ex.task_execution.name
    #                  if action_ex.task_execution else None)
    #     setattr(res, 'task_name', task_name)

    return res


def _get_action_executions(task_execution_id=None, marker=None, limit=None,
                           sort_keys='created_at', sort_dirs='asc',
                           fields='', include_output=False, **filters):
    """Return all action executions.

    Where project_id is the same as the requester or
    project_id is different but the scope is public.

    :param marker: Optional. Pagination marker for large data sets.
    :param limit: Optional. Maximum number of resources to return in a
                  single result. Default value is None for backward
                  compatibility.
    :param sort_keys: Optional. Columns to sort results by.
                      Default: created_at, which is backward compatible.
    :param sort_dirs: Optional. Directions to sort corresponding to
                      sort_keys, "asc" or "desc" can be chosen.
                      Default: desc. The length of sort_dirs can be equal
                      or less than that of sort_keys.
    :param fields: Optional. A specified list of fields of the resource to
                   be returned. 'id' will be included automatically in
                   fields if it's not provided, since it will be used when
                   constructing 'next' link.
    :param filters: Optional. A list of filters to apply to the result.
    """
    if task_execution_id:
        filters['task_execution_id'] = task_execution_id

    if include_output:
        resource_function = _get_action_execution_resource
    else:
        resource_function = _get_action_execution_resource_for_list

    return rest_utils.get_all(
        resources.ActionExecutions,
        resources.ActionExecution,
        db_api.get_action_executions,
        db_api.get_action_execution,
        resource_function=resource_function,
        marker=marker,
        limit=limit,
        sort_keys=sort_keys,
        sort_dirs=sort_dirs,
        fields=fields,
        **filters
    )


class ActionExecutionsController(rest.RestController):
    @rest_utils.wrap_wsme_controller_exception
    @wsme_pecan.wsexpose(resources.ActionExecution, wtypes.text,
                         types.uniquelist)
    def get(self, id, fields=None):
        """Return the specified action_execution.

        :param id: UUID of action execution to retrieve
        :param fields: Optional. A specified list of fields of the resource to
                       be returned. 'id' will be included automatically in
                       fields if it's not provided.
        """
        acl.enforce('action_executions:get', context.ctx())

        LOG.debug("Fetch action_execution [id=%s]", id)

        return _get_action_execution(id, fields=fields)

    @rest_utils.wrap_wsme_controller_exception
    @wsme_pecan.wsexpose(resources.ActionExecution,
                         body=resources.ActionExecution, status_code=201)
    def post(self, action_ex):
        """Create new action_execution.

        :param action_ex: Action to execute
        """
        acl.enforce('action_executions:create', context.ctx())
        LOG.debug(
            "Create action_execution [action_execution=%s]",
            action_ex
        )
        name = action_ex.name
        description = action_ex.description or None
        action_input = action_ex.input or {}
        params = action_ex.params or {}
        namespace = action_ex.workflow_namespace or ''

        if not name:
            raise exc.InputException(
                "Please provide at least action name to run action."
            )

        values = rpc.get_engine_client().start_action(
            name,
            action_input,
            description=description,
            namespace=namespace,
            **params
        )

        return resources.ActionExecution.from_dict(values)

    @rest_utils.wrap_wsme_controller_exception
    @wsme_pecan.wsexpose(
        resources.ActionExecution,
        wtypes.text,
        body=resources.ActionExecution
    )
    def put(self, id, action_ex):
        """Update the specified action_execution.

        :param id: UUID of action execution to update
        :param action_ex: Action execution for update
        """
        acl.enforce('action_executions:update', context.ctx())

        LOG.debug(
            "Update action_execution [id=%s, action_execution=%s]",
            id,
            action_ex
        )

        if action_ex.state not in SUPPORTED_TRANSITION_STATES:
            raise exc.InvalidResultException(
                "Error. Expected one of %s, actual: %s" % (
                    SUPPORTED_TRANSITION_STATES,
                    action_ex.state
                )
            )

        if states.is_completed(action_ex.state):
            output = action_ex.output

            if action_ex.state == states.SUCCESS:
                result = ml_actions.Result(data=output)
            elif action_ex.state == states.ERROR:
                if not output:
                    output = 'Unknown error'
                result = ml_actions.Result(error=output)
            elif action_ex.state == states.CANCELLED:
                result = ml_actions.Result(cancel=True)

            values = rpc.get_engine_client().on_action_complete(id, result)

        if action_ex.state in [states.PAUSED, states.RUNNING]:
            state = action_ex.state
            values = rpc.get_engine_client().on_action_update(id, state)

        return resources.ActionExecution.from_dict(values)

    @rest_utils.wrap_wsme_controller_exception
    @wsme_pecan.wsexpose(resources.ActionExecutions, types.uuid, int,
                         types.uniquelist, types.list, types.uniquelist,
                         wtypes.text, wtypes.text, wtypes.text,
                         wtypes.text, wtypes.text, wtypes.text, types.uuid,
                         wtypes.text, wtypes.text, bool, types.jsontype,
                         types.jsontype, types.jsontype, wtypes.text, bool)
    def get_all(self, marker=None, limit=None, sort_keys='created_at',
                sort_dirs='asc', fields='', created_at=None, name=None,
                tags=None, updated_at=None, workflow_name=None,
                task_name=None, task_execution_id=None, state=None,
                state_info=None, accepted=None, input=None, output=None,
                params=None, description=None, include_output=False):
        """Return all tasks within the execution.

        Where project_id is the same as the requester or
        project_id is different but the scope is public.

        :param marker: Optional. Pagination marker for large data sets.
        :param limit: Optional. Maximum number of resources to return in a
                      single result. Default value is None for backward
                      compatibility.
        :param sort_keys: Optional. Columns to sort results by.
                          Default: created_at, which is backward compatible.
        :param sort_dirs: Optional. Directions to sort corresponding to
                          sort_keys, "asc" or "desc" can be chosen.
                          Default: desc. The length of sort_dirs can be equal
                          or less than that of sort_keys.
        :param fields: Optional. A specified list of fields of the resource to
                       be returned. 'id' will be included automatically in
                       fields if it's not provided, since it will be used when
                       constructing 'next' link.
        :param name: Optional. Keep only resources with a specific name.
        :param workflow_name: Optional. Keep only resources with a specific
                              workflow name.
        :param task_name: Optional. Keep only resources with a specific
                          task name.
        :param task_execution_id: Optional. Keep only resources within a
                                  specific task execution.
        :param state: Optional. Keep only resources with a specific state.
        :param state_info: Optional. Keep only resources with specific state
                           information.
        :param accepted: Optional. Keep only resources which have been accepted
                         or not.
        :param input: Optional. Keep only resources with a specific input.
        :param output: Optional. Keep only resources with a specific output.
        :param params: Optional. Keep only resources with specific parameters.
        :param description: Optional. Keep only resources with a specific
                            description.
        :param tags: Optional. Keep only resources containing specific tags.
        :param created_at: Optional. Keep only resources created at a specific
                           time and date.
        :param updated_at: Optional. Keep only resources with specific latest
                           update time and date.
        :param include_output: Optional. Include the output for all executions
                               in the list
        """
        acl.enforce('action_executions:list', context.ctx())

        filters = filter_utils.create_filters_from_request_params(
            created_at=created_at,
            name=name,
            tags=tags,
            updated_at=updated_at,
            workflow_name=workflow_name,
            task_name=task_name,
            task_execution_id=task_execution_id,
            state=state,
            state_info=state_info,
            accepted=accepted,
            input=input,
            output=output,
            params=params,
            description=description
        )

        LOG.debug(
            "Fetch action_executions. marker=%s, limit=%s, "
            "sort_keys=%s, sort_dirs=%s, filters=%s",
            marker,
            limit,
            sort_keys,
            sort_dirs,
            filters
        )

        return _get_action_executions(
            marker=marker,
            limit=limit,
            sort_keys=sort_keys,
            sort_dirs=sort_dirs,
            fields=fields,
            include_output=include_output,
            **filters
        )

    @rest_utils.wrap_wsme_controller_exception
    @wsme_pecan.wsexpose(None, wtypes.text, status_code=204)
    def delete(self, id):
        """Delete the specified action_execution.

        :param id: UUID of action execution to delete
        """
        acl.enforce('action_executions:delete', context.ctx())

        LOG.debug("Delete action_execution [id=%s]", id)

        if not cfg.CONF.api.allow_action_execution_deletion:
            raise exc.NotAllowedException("Action execution deletion is not "
                                          "allowed.")

        with db_api.transaction():
            action_ex = db_api.get_action_execution(id)

            if action_ex.task_execution_id:
                raise exc.NotAllowedException(
                    "Only ad-hoc action execution can be deleted."
                )

            if not states.is_completed(action_ex.state):
                raise exc.NotAllowedException(
                    "Only completed action execution can be deleted."
                )

            return db_api.delete_action_execution(id)


class TasksActionExecutionController(rest.RestController):
    @rest_utils.wrap_wsme_controller_exception
    @wsme_pecan.wsexpose(resources.ActionExecutions, types.uuid, types.uuid,
                         int, types.uniquelist, types.list, types.uniquelist,
                         wtypes.text, types.uniquelist, wtypes.text,
                         wtypes.text, wtypes.text, wtypes.text, wtypes.text,
                         wtypes.text, bool, types.jsontype, types.jsontype,
                         types.jsontype, wtypes.text, bool)
    def get_all(self, task_execution_id, marker=None, limit=None,
                sort_keys='created_at', sort_dirs='asc', fields='',
                created_at=None, name=None, tags=None,
                updated_at=None, workflow_name=None, task_name=None,
                state=None, state_info=None, accepted=None, input=None,
                output=None, params=None, description=None,
                include_output=None):
        """Return all tasks within the execution.

        Where project_id is the same as the requester or
        project_id is different but the scope is public.

        :param task_execution_id: Keep only resources within a specific task
                                  execution.
        :param marker: Optional. Pagination marker for large data sets.
        :param limit: Optional. Maximum number of resources to return in a
                      single result. Default value is None for backward
                      compatibility.
        :param sort_keys: Optional. Columns to sort results by.
                          Default: created_at, which is backward compatible.
        :param sort_dirs: Optional. Directions to sort corresponding to
                          sort_keys, "asc" or "desc" can be chosen.
                          Default: desc. The length of sort_dirs can be equal
                          or less than that of sort_keys.
        :param fields: Optional. A specified list of fields of the resource to
                       be returned. 'id' will be included automatically in
                       fields if it's not provided, since it will be used when
                       constructing 'next' link.
        :param name: Optional. Keep only resources with a specific name.
        :param workflow_name: Optional. Keep only resources with a specific
                              workflow name.
        :param task_name: Optional. Keep only resources with a specific
                          task name.
        :param state: Optional. Keep only resources with a specific state.
        :param state_info: Optional. Keep only resources with specific state
                           information.
        :param accepted: Optional. Keep only resources which have been accepted
                         or not.
        :param input: Optional. Keep only resources with a specific input.
        :param output: Optional. Keep only resources with a specific output.
        :param params: Optional. Keep only resources with specific parameters.
        :param description: Optional. Keep only resources with a specific
                            description.
        :param tags: Optional. Keep only resources containing specific tags.
        :param created_at: Optional. Keep only resources created at a specific
                           time and date.
        :param updated_at: Optional. Keep only resources with specific latest
                           update time and date.
        :param include_output: Optional. Include the output for all executions
                               in the list
        """
        acl.enforce('action_executions:list', context.ctx())

        filters = filter_utils.create_filters_from_request_params(
            created_at=created_at,
            name=name,
            tags=tags,
            updated_at=updated_at,
            workflow_name=workflow_name,
            task_name=task_name,
            task_execution_id=task_execution_id,
            state=state,
            state_info=state_info,
            accepted=accepted,
            input=input,
            output=output,
            params=params,
            description=description
        )

        LOG.debug(
            "Fetch action_executions. marker=%s, limit=%s, "
            "sort_keys=%s, sort_dirs=%s, filters=%s",
            marker,
            limit,
            sort_keys,
            sort_dirs,
            filters
        )

        return _get_action_executions(
            marker=marker,
            limit=limit,
            sort_keys=sort_keys,
            sort_dirs=sort_dirs,
            fields=fields,
            include_output=include_output,
            **filters
        )

    @rest_utils.wrap_wsme_controller_exception
    @wsme_pecan.wsexpose(resources.ActionExecution, wtypes.text, wtypes.text,
                         types.uniquelist)
    def get(self, task_execution_id, action_ex_id, fields=()):
        """Return the specified action_execution.

        :param task_execution_id: Task execution UUID
        :param action_ex_id: Action execution UUID
        :param fields: Optional. A specified list of fields of the resource to
                       be returned. 'id' will be included automatically in
                       fields if it's not provided.
        """
        acl.enforce('action_executions:get', context.ctx())

        LOG.debug("Fetch action_execution [id=%s]", action_ex_id)

        return _get_action_execution(action_ex_id, fields=fields)



#  Copyright 2019 - Nokia Corporation
#
#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

import yaml
from yaml import *  # noqa

yaml.SafeDumper.ignore_aliases = lambda *args: True


class SafeLoader(yaml.SafeLoader):
    """Treat '@', '&', '*' as plain string.

       Anchors are not used in mistral workflow. It's better to
       disable them completely. Anchors can be used as an exploit to a
       Denial of service attack through expansion (Billion Laughs)
       see https://en.wikipedia.org/wiki/Billion_laughs_attack.
       Also this module uses the safe loader by default which is always
       a better loader.

       When using yaml module to load a yaml file or a string use this
       module instead of yaml.

       Example:
         import mistral.utils.safe_yaml as safe_yaml
         ...
         ...

         safe_yaml.load(...)

    """

    def fetch_alias(self):
        return self.fetch_plain()

    def fetch_anchor(self):
        return self.fetch_plain()

    def check_plain(self):
        # Modified: allow '@'
        if self.peek() == '@':
            return True
        else:
            return super(SafeLoader, self).check_plain()


def load(stream):
    return yaml.load(stream, SafeLoader)


def safe_load(stream):
    return load(stream)



# Copyright 2020 Nokia Software.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""
This module is a facade for all action subsystem. It represents a
collection of functions for accessing information about actions
available in the system.
"""

from oslo_log import log as logging
from stevedore import extension

from mistral_lib import actions as ml_actions

from mistral.actions import test

LOG = logging.getLogger(__name__)

_SYSTEM_PROVIDER = None
_TEST_PROVIDER = None


def _get_registered_providers():
    providers = []

    mgr = extension.ExtensionManager(
        namespace='mistral.action.providers',
        invoke_on_load=False
    )

    for provider_name in mgr.names():
        provider_cls = mgr[provider_name].plugin

        try:
            providers.append(provider_cls(provider_name))
        except Exception:
            LOG.exception(
                'Failed to instantiate an action provider from the class: %s',
                provider_cls
            )

            raise

    if not providers:
        LOG.warning("No action providers found in the system.")

    return providers


def get_test_action_provider():
    """Returns a singleton for the test action provider."""

    global _TEST_PROVIDER

    if _TEST_PROVIDER is None:
        _TEST_PROVIDER = test.TestActionProvider()

    return _TEST_PROVIDER


def get_system_action_provider():
    """Returns a singleton for the system action provider.

    In fact, this method serves a facade for the entire action subsystem.
    Clients of the acton subsystem must get the system action provider
    and work with actions through it. The system action provider created
    by this method (on the first call) is nothing but just a composite
    on top of the action providers registered in the entry point
    "mistral.action.providers".
    """

    global _SYSTEM_PROVIDER

    if _SYSTEM_PROVIDER is None:
        delegates = _get_registered_providers()

        # Add an action provider for testing to the end of the list
        # so that it has the lowest priority. In production runs it's
        # always empty so it won't take any effect.
        delegates.append(get_test_action_provider())

        _SYSTEM_PROVIDER = ml_actions.CompositeActionProvider(
            'system',
            delegates
        )

    return _SYSTEM_PROVIDER



# Copyright 2020 - Nokia Networks.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from oslo_log import log as logging
from pecan import request
from pecan import rest
import wsmeext.pecan as wsme_pecan

from mistral.api.controllers.v2 import resources
from mistral.api.controllers.v2 import types
from mistral.db.v2 import api as db_api
from mistral.utils import rest_utils
from mistral.workflow import states

LOG = logging.getLogger(__name__)


def get_task_sub_executions_list(task_ex_id, filters, cur_depth):
    task_sub_execs = []

    with db_api.transaction():
        task_ex = db_api.get_task_execution(task_ex_id)

        if filters['errors_only'] and task_ex.state != states.ERROR:
            return []

        child_wf_executions = task_ex.workflow_executions

    for c_ex in child_wf_executions:
        task_sub_execs.extend(
            get_execution_sub_executions_list(
                c_ex.id,
                filters,
                cur_depth
            )
        )

    return task_sub_execs


def get_execution_sub_executions_list(wf_ex_id, filters, cur_depth):
    max_depth = filters['max_depth']
    include_output = filters['include_output']
    ex_sub_execs = []

    if 0 <= max_depth < cur_depth:
        return []

    with db_api.transaction():
        wf_ex = db_api.get_workflow_execution(wf_ex_id)

        wf_resource = _get_wf_resource_from_db_model(
            wf_ex,
            include_output)

        ex_sub_execs.append(wf_resource)

        task_execs = wf_ex.task_executions

    for t_ex in task_execs:
        task_sub_executions = get_task_sub_executions_list(
            t_ex.id,
            filters,
            cur_depth + 1
        )
        ex_sub_execs.extend(task_sub_executions)

    return ex_sub_execs


def _get_wf_resource_from_db_model(wf_ex, include_output):
    if include_output:
        rest_utils.load_deferred_fields(wf_ex, ['params', 'input', 'output'])
    else:
        rest_utils.load_deferred_fields(wf_ex, ['params', 'input'])

    return resources.Execution.from_db_model(wf_ex)


def _get_sub_executions(origin, id, filters):
    if origin == 'execution':
        return get_execution_sub_executions_list(id, filters, cur_depth=0)
    else:
        return get_task_sub_executions_list(id, filters, cur_depth=0)


class SubExecutionsController(rest.RestController):
    @rest_utils.wrap_wsme_controller_exception
    @wsme_pecan.wsexpose(resources.Executions, types.uuid, bool, int, bool)
    def get(self, id, errors_only=False, max_depth=-1, include_output=False):
        """Return workflow execution report.

        :param id: The ID of the workflow execution or task execution
            to get the sub-executions of.
        :param errors_only: Optional. If True, only error paths of the
            execution tree are returned .
        :param max_depth: Optional. Limits the depth of recursion while
            obtaining the execution tree. If a value of the
            flag is a negative number then no limit is set.
        :param include_output: Optional. Include the output for all executions
            in the list.
        """
        origin = 'execution' if request.path.startswith('/v2/executions') \
            else 'task'

        LOG.info(
            "Fetching sub executions of %s [id=%s]",
            origin,
            id
        )

        filters = {
            'errors_only': errors_only,
            'max_depth': max_depth,
            'include_output': include_output
        }

        sub_executions_resource = _get_sub_executions(origin, id, filters)

        return resources.Executions.convert_with_links(
            sub_executions_resource,
            request.application_url,
        )



# Copyright 2013 - Mirantis, Inc.
# Copyright 2015 - StackStorm, Inc.
# Copyright 2015 - Huawei Technologies Co. Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from mistral import exceptions as exc
from mistral.tests.unit import base
from mistral.utils import ssh_utils
from mistral_lib import utils


class UtilsTest(base.BaseTest):

    def test_itersubclasses(self):
        class A(object):
            pass

        class B(A):
            pass

        class C(A):
            pass

        class D(C):
            pass

        self.assertEqual([B, C, D], list(utils.iter_subclasses(A)))

    def test_paramiko_to_private_key(self):
        self.assertRaises(
            exc.DataAccessException,
            ssh_utils._to_paramiko_private_key,
            "../dir"
        )
        self.assertRaises(
            exc.DataAccessException,
            ssh_utils._to_paramiko_private_key,
            "..\\dir"
        )

        self.assertIsNone(
            ssh_utils._to_paramiko_private_key(private_key_filename=None,
                                               password='pass')
        )



# Copyright 2016 - Nokia Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from cachetools import keys as cachetools_keys
import decorator
import functools
import inspect

from alembic import op
from sqlalchemy import exc as sqla_exc
from sqlalchemy import inspect as ins

from oslo_db import exception as db_exc
from oslo_log import log as logging

import tenacity

from mistral import context
from mistral.db.sqlalchemy import base as db_base
from mistral import exceptions as exc
from mistral.services import security
from mistral_lib import utils as ml_utils


LOG = logging.getLogger(__name__)

_RETRY_ERRORS = (
    db_exc.DBDeadlock,
    db_exc.DBConnectionError,
    sqla_exc.OperationalError
)


def _with_auth_context(auth_ctx, func, *args, **kw):
    """Runs the given function with the specified auth context.

    :param auth_ctx: Authentication context.
    :param func: Function to run with the specified auth context.
    :param args: Function positional arguments.
    :param kw: Function keyword arguments.
    :return: Function result.
    """
    old_auth_ctx = context.ctx() if context.has_ctx() else None

    context.set_ctx(auth_ctx)

    try:
        return func(*args, **kw)
    except Exception as e:
        # Note (rakhmerov): In case of "Too many connections" error from the
        # database it doesn't get wrapped with a SQLAlchemy exception for some
        # reason so we have to check the exception message explicitly.
        if isinstance(e, _RETRY_ERRORS) or 'Too many connections' in str(e):
            LOG.exception(
                "DB error detected, operation will be retried: %s", func
            )

        raise
    finally:
        context.set_ctx(old_auth_ctx)


def retry_on_db_error(func, retry=None):
    """Decorates the given function so that it retries on DB errors.

    Note that the decorator retries the function/method only on some
    of the DB errors that are considered to be worth retrying, like
    deadlocks and disconnections.

    :param func: Function to decorate.
    :param retry: a Retrying object
    :return: Decorated function.
    """
    if not retry:
        retry = tenacity.Retrying(
            retry=(
                tenacity.retry_if_exception_type(_RETRY_ERRORS) |
                tenacity.retry_if_exception_message(
                    match='Too many connections'
                )
            ),
            stop=tenacity.stop_after_attempt(50),
            wait=tenacity.wait_incrementing(start=0, increment=0.1, max=2)
        )

    # The `assigned` arg should be empty as some of the default values are not
    # supported by simply initialized MagicMocks. The consequence may
    # be that the representation will contain the wrapper and not the
    # wrapped function.
    @functools.wraps(func, assigned=[])
    def decorate(*args, **kw):
        # Retrying library decorator might potentially run a decorated
        # function within a new thread so it's safer not to apply the
        # decorator directly to a target method/function because we can
        # lose an authentication context.
        # The solution is to create one more function and explicitly set
        # auth context before calling it (potentially in a new thread).
        auth_ctx = context.ctx() if context.has_ctx() else None

        return retry(_with_auth_context, auth_ctx, func, *args, **kw)

    return decorate


def check_db_obj_access(db_obj):
    """Check accessibility to db object."""
    ctx = context.ctx()
    is_admin = ctx.is_admin

    if not is_admin and db_obj.project_id != security.get_project_id():
        raise exc.NotAllowedException(
            "Can not access %s resource of other projects, ID: %s" %
            (db_obj.__class__.__name__, db_obj.id)
        )

    if not is_admin and hasattr(db_obj, 'is_system') and db_obj.is_system:
        raise exc.InvalidActionException(
            "Can not modify a system %s resource, ID: %s" %
            (db_obj.__class__.__name__, db_obj.id)
        )


def tx_cached(use_args=None, ignore_args=None):
    """Decorates any function to cache its result within a DB transaction.

    Since a DB transaction is coupled with the current thread, the scope
    of the underlying cache doesn't go beyond the thread. The decorator
    is mainly useful for situations when we know we can safely cache a
    result of some calculation if we know that it's not going to change
    till the end of the current transaction.

    :param use_args: A tuple with argument names of the decorated function
        used to build a cache key.
    :param ignore_args:  A tuple with argument names of the decorated function
        that should be ignored when building a cache key.
    :return: Decorated function.
    """

    if use_args and ignore_args:
        raise ValueError(
            "Only one of 'use_args' and 'ignore_args' can be used."
        )

    def _build_cache_key(func, *args, **kw):
        # { arg name => arg value }
        arg_dict = inspect.getcallargs(func, *args, **kw)

        if ignore_args:
            if not isinstance(ignore_args, (str, tuple)):
                raise ValueError(
                    "'ignore_args' must be either a tuple or a string,"
                    " actual type: %s" % type(ignore_args)
                )

            ignore_args_tup = (
                ignore_args if isinstance(ignore_args, tuple) else
                (ignore_args,)
            )

            for arg_name in ignore_args_tup:
                arg_dict.pop(arg_name, None)

        if use_args:
            if not isinstance(use_args, (str, tuple)):
                raise ValueError(
                    "'use_args' must be either a tuple or a string,"
                    " actual type: %s" % type(use_args)
                )

            use_args_tup = (
                use_args if isinstance(use_args, tuple) else (use_args,)
            )

            for arg_name in arg_dict.keys():
                if arg_name not in tuple(use_args_tup):
                    arg_dict.pop(arg_name, None)

        return cachetools_keys.hashkey(**arg_dict)

    @decorator.decorator
    def _decorator(func, *args, **kw):
        cache = db_base.get_tx_scoped_cache()

        # A DB transaction may not be necessarily open at the moment.
        if cache is None:
            return func(*args, **kw)

        cache_key = _build_cache_key(func, *args, **kw)

        result = cache.get(cache_key, default=ml_utils.NotDefined)

        if result is not ml_utils.NotDefined:
            return result

        # We don't do any exception handling here. In case of an exception
        # nothing will be put into the cache and the exception will just
        # bubble up as if there wasn't any wrapper.
        result = func(*args, **kw)

        cache[cache_key] = result

        return result

    return _decorator


def column_exists(table_name, column_name):
    bind = op.get_context().bind
    insp = ins(bind)
    columns = insp.get_columns(table_name)
    return any(c["name"] == column_name for c in columns)



#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import eventlet

from mistral.tests.unit.engine import base

from mistral.db.v2 import api as db_api
from mistral.services import workflows as wf_service


class TaskStartedFinishedAtTest(base.EngineTestCase):
    def setUp(self):
        super(TaskStartedFinishedAtTest, self).setUp()

    def test_started_finished_fields_updated_after_rerun(self):
        wf_text = """
        version: '2.0'

        wf:
          tasks:
            task1:
              action: std.fail
              wait-before: 2
        """

        wf_service.create_workflows(wf_text)

        wf_ex = self.engine.start_workflow('wf')

        self.await_workflow_error(wf_ex.id)

        task_ex = self._extract_task_ex(wf_ex.id)

        started_1st, finished_1st = self._get_started_finished(task_ex)

        # Make sure to rerun the workflow after a certain delay so that
        # times for the first run are different from times in the second run.
        eventlet.sleep(1)

        wf_ex = self.engine.rerun_workflow(task_ex.id)

        self.await_workflow_error(wf_ex.id)

        task_ex = self._extract_task_ex(wf_ex.id)

        started_2nd, finished_2nd = self._get_started_finished(task_ex)

        self.assertNotEqual(started_1st, started_2nd)
        self.assertNotEqual(finished_1st, finished_2nd)

    def test_correct_duration_in_case_of_join_all(self):
        wf_text = """
        version: '2.0'

        wf:
          tasks:
            task1:
              action: std.sleep seconds=1
              on-success: join_task

            task2:
              action: std.sleep seconds=2
              on-success: join_task

            task3:
              action: std.sleep seconds=3
              on-success: join_task

            join_task:
              join: all
              action: std.sleep seconds=1
        """

        wf_service.create_workflows(wf_text)

        wf_ex = self.engine.start_workflow('wf')

        self.await_workflow_success(wf_ex.id)

        task1_ex = self._extract_task_ex(wf_ex.id, 'task1')
        task2_ex = self._extract_task_ex(wf_ex.id, 'task2')
        task3_ex = self._extract_task_ex(wf_ex.id, 'task3')

        join_task_ex = self._extract_task_ex(wf_ex.id, 'join_task')

        self._check_started_after(join_task_ex, task1_ex)
        self._check_started_after(join_task_ex, task2_ex)
        self._check_started_after(join_task_ex, task3_ex)

    def test_retries_do_not_update_created_at(self):
        wf_text = """
        version: '2.0'

        wf:
          tasks:
            task1:
              action: std.fail
              retry:
                delay: 1
                count: 3
        """

        wf_service.create_workflows(wf_text)

        wf_ex = self.engine.start_workflow('wf')

        task_ex = self._extract_task_ex(wf_ex.id)
        created_at_before_retry = task_ex.created_at

        self.await_workflow_error(wf_ex.id)

        task_ex = self._extract_task_ex(wf_ex.id)
        created_at_after_retry = task_ex.created_at

        self.assertEqual(created_at_before_retry, created_at_after_retry)

    def test_wait_before_after_are_included_to_duration(self):
        wf_text = """
        version: '2.0'

        wf:
          tasks:
            task1:
              action: std.noop
              wait-before: 1
              wait-after: 2
        """

        wf_service.create_workflows(wf_text)

        wf_ex = self.engine.start_workflow('wf')

        self.await_workflow_success(wf_ex.id)

        task_ex = self._extract_task_ex(wf_ex.id)

        started, finished = self._get_started_finished(task_ex)

        duration = self._get_task_duration(started, finished)

        self._check_duration_more_than(duration, 1)

    def _extract_task_ex(self, wf_ex_id, name='task1'):
        with db_api.transaction():
            wf_ex = db_api.get_workflow_execution(wf_ex_id)

            task_execs = wf_ex.task_executions

        return self._assert_single_item(task_execs, name=name)

    def _get_started_finished(self, task_ex):
        started_at = task_ex.started_at
        finished_at = task_ex.finished_at

        self.assertIsNotNone(started_at)
        self.assertIsNotNone(finished_at)

        return started_at, finished_at

    def _get_task_duration(self, start_time, finish_time):
        return (finish_time - start_time).total_seconds()

    def _check_started_after(self, task_ex1, task_ex2):
        first_finished = self._get_started_finished(task_ex2)[1]
        second_started = self._get_started_finished(task_ex1)[0]

        delta = self._get_task_duration(first_finished, second_started)

        self.assertTrue(
            delta >= 0,
            "Expected {} was started after {} was finished".format(
                task_ex1.name, task_ex2.name)
        )

    def _check_duration_more_than(self, duration, time):
        self.assertTrue(
            time < duration,
            "Expected duration {} was more than {}".format(duration, time)
        )



# Copyright 2015 - Mirantis, Inc.
# Copyright 2015 - StackStorm, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.


from oslo_config import cfg

from mistral.db.v2 import api as db_api
from mistral.tests.unit import base as test_base


WF_EXECS = [
    {
        'name': '1',
        'spec': {},
        'start_params': {},
        'state': 'RUNNING',
        'state_info': "Running...",
        'created_at': None,
        'updated_at': None,
        'context': None,
        'task_id': None,
        'trust_id': None
    },
    {
        'name': '1',
        'spec': {},
        'start_params': {},
        'state': 'RUNNING',
        'state_info': "Running...",
        'created_at': None,
        'updated_at': None,
        'context': None,
        'task_id': None,
        'trust_id': None
    }
]


class TransactionsTest(test_base.DbTestCase):

    def setUp(self):
        super(TransactionsTest, self).setUp()

        cfg.CONF.set_default('auth_enable', True, group='pecan')

        self.addCleanup(
            cfg.CONF.set_default,
            'auth_enable',
            False,
            group='pecan'
        )

    def test_read_only_transactions(self):
        with db_api.transaction():
            db_api.create_workflow_execution(WF_EXECS[0])

            wf_execs = db_api.get_workflow_executions()
            self.assertEqual(1, len(wf_execs))

        wf_execs = db_api.get_workflow_executions()
        self.assertEqual(1, len(wf_execs))

        with db_api.transaction(read_only=True):
            db_api.create_workflow_execution(WF_EXECS[1])

            wf_execs = db_api.get_workflow_executions()
            self.assertEqual(2, len(wf_execs))

        wf_execs = db_api.get_workflow_executions()
        self.assertEqual(1, len(wf_execs))



# Copyright 2018 - Nokia, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.


from mistral.tests.unit import base
from mistral.utils import filter_utils


class FilterUtilsTest(base.BaseTest):
    def test_create_filters_with_nones(self):
        expected_filters = {
            'key2': {'eq': 'value2'},
            'key1': {'eq': None}
        }

        filters = filter_utils.create_filters_from_request_params(
            none_values=['key1'],
            key1=None,
            key2='value2',
            key3=None,
        )

        self.assertEqual(expected_filters, filters)

        del expected_filters['key1']

        filters = filter_utils.create_filters_from_request_params(
            none_values=[],
            key1=None,
            key2='value2',
            key3=None,
        )

        self.assertEqual(expected_filters, filters)



# Copyright 2016 Catalyst IT Ltd
# Copyright 2017 Brocade Communications Systems, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import copy
import time
from unittest import mock

from oslo_config import cfg

from mistral import context as auth_context
from mistral.db.v2.sqlalchemy import api as db_api
from mistral.event_engine import default_event_engine as evt_eng
from mistral.rpc import clients as rpc
from mistral.services import workflows
from mistral.tests.unit import base

WORKFLOW_LIST = """
---
version: '2.0'

my_wf:
  type: direct

  tasks:
    task1:
      action: std.echo output='Hi!'
"""

EXCHANGE_TOPIC = ('openstack', 'notification')
EVENT_TYPE = 'compute.instance.create.start'

EVENT_TRIGGER = {
    'name': 'trigger1',
    'workflow_id': '',
    'workflow_input': {},
    'workflow_params': {},
    'exchange': 'openstack',
    'topic': 'notification',
    'event': EVENT_TYPE,
}

cfg.CONF.set_default('auth_enable', False, group='pecan')


class EventEngineTest(base.DbTestCase):
    def setUp(self):
        super(EventEngineTest, self).setUp()

        self.wf = workflows.create_workflows(WORKFLOW_LIST)[0]

        EVENT_TRIGGER['workflow_id'] = self.wf.id

    @mock.patch.object(rpc, 'get_engine_client', mock.Mock())
    def test_event_engine_start_with_no_triggers(self):
        e_engine = evt_eng.DefaultEventEngine()

        self.addCleanup(e_engine.handler_tg.stop)

        self.assertEqual(0, len(e_engine.event_triggers_map))
        self.assertEqual(0, len(e_engine.exchange_topic_events_map))
        self.assertEqual(0, len(e_engine.exchange_topic_listener_map))

    @mock.patch('mistral.messaging.start_listener')
    @mock.patch.object(rpc, 'get_engine_client', mock.Mock())
    def test_event_engine_start_with_triggers(self, mock_start):
        trigger = db_api.create_event_trigger(EVENT_TRIGGER)

        e_engine = evt_eng.DefaultEventEngine()

        self.addCleanup(e_engine.handler_tg.stop)

        self.assertEqual(1, len(e_engine.exchange_topic_events_map))
        self.assertEqual(
            EVENT_TYPE,
            list(e_engine.exchange_topic_events_map[EXCHANGE_TOPIC])[0]
        )
        self.assertEqual(1, len(e_engine.event_triggers_map))
        self.assertEqual(1, len(e_engine.event_triggers_map[EVENT_TYPE]))
        self._assert_dict_contains_subset(
            trigger.to_dict(),
            e_engine.event_triggers_map[EVENT_TYPE][0]
        )
        self.assertEqual(1, len(e_engine.exchange_topic_listener_map))

    @mock.patch('mistral.messaging.start_listener')
    @mock.patch.object(rpc, 'get_engine_client', mock.Mock())
    def test_event_engine_public_trigger(self, mock_start):
        t = copy.deepcopy(EVENT_TRIGGER)

        # Create public trigger as an admin
        self.ctx = base.get_context(default=False, admin=True)
        auth_context.set_ctx(self.ctx)

        t['scope'] = 'public'
        t['project_id'] = self.ctx.project_id
        trigger = db_api.create_event_trigger(t)

        # Switch to the user.
        self.ctx = base.get_context(default=True)
        auth_context.set_ctx(self.ctx)

        e_engine = evt_eng.DefaultEventEngine()

        self.addCleanup(e_engine.handler_tg.stop)

        event = {
            'event_type': EVENT_TYPE,
            'payload': {},
            'publisher': 'fake_publisher',
            'timestamp': '',
            'context': {
                'project_id': '%s' % self.ctx.project_id,
                'user_id': 'fake_user'
            },
        }

        # Moreover, assert that trigger.project_id != event.project_id
        self.assertNotEqual(
            trigger.project_id, event['context']['project_id']
        )

        with mock.patch.object(e_engine, 'engine_client') as client_mock:
            e_engine.event_queue.put(event)

            time.sleep(1)

            self.assertEqual(1, client_mock.start_workflow.call_count)

            args, kwargs = client_mock.start_workflow.call_args

            self.assertEqual(
                (EVENT_TRIGGER['workflow_id'], '', None, {}),
                args
            )
            self.assertDictEqual(
                {
                    'service': 'fake_publisher',
                    'project_id': '%s' % self.ctx.project_id,
                    'user_id': 'fake_user',
                    'timestamp': ''
                },
                kwargs['event_params']
            )

    @mock.patch('mistral.messaging.start_listener')
    @mock.patch.object(rpc, 'get_engine_client', mock.Mock())
    def test_process_event_queue(self, mock_start):
        EVENT_TRIGGER['project_id'] = self.ctx.project_id
        db_api.create_event_trigger(EVENT_TRIGGER)

        e_engine = evt_eng.DefaultEventEngine()

        self.addCleanup(e_engine.handler_tg.stop)

        event = {
            'event_type': EVENT_TYPE,
            'payload': {},
            'publisher': 'fake_publisher',
            'timestamp': '',
            'context': {
                'project_id': '%s' % self.ctx.project_id,
                'user_id': 'fake_user'
            },
        }

        with mock.patch.object(e_engine, 'engine_client') as client_mock:
            e_engine.event_queue.put(event)

            time.sleep(1)

            self.assertEqual(1, client_mock.start_workflow.call_count)

            args, kwargs = client_mock.start_workflow.call_args

            self.assertEqual(
                (EVENT_TRIGGER['workflow_id'], '', None, {}),
                args
            )
            self.assertDictEqual(
                {
                    'service': 'fake_publisher',
                    'project_id': '%s' % self.ctx.project_id,
                    'user_id': 'fake_user',
                    'timestamp': ''
                },
                kwargs['event_params']
            )


class NotificationsConverterTest(base.BaseTest):
    def test_convert(self):
        definition_cfg = [
            {
                'event_types': EVENT_TYPE,
                'properties': {'resource_id': '<% $.payload.instance_id %>'}
            }
        ]

        converter = evt_eng.NotificationsConverter()
        converter.definitions = [evt_eng.EventDefinition(event_def)
                                 for event_def in reversed(definition_cfg)]

        notification = {
            'event_type': EVENT_TYPE,
            'payload': {'instance_id': '12345'},
            'publisher': 'fake_publisher',
            'timestamp': '',
            'context': {'project_id': 'fake_project', 'user_id': 'fake_user'}
        }

        event = converter.convert(EVENT_TYPE, notification)

        self.assertDictEqual(
            {'resource_id': '12345'},
            event
        )

    def test_convert_event_type_not_defined(self):
        definition_cfg = [
            {
                'event_types': EVENT_TYPE,
                'properties': {'resource_id': '<% $.payload.instance_id %>'}
            }
        ]

        converter = evt_eng.NotificationsConverter()
        converter.definitions = [evt_eng.EventDefinition(event_def)
                                 for event_def in reversed(definition_cfg)]

        notification = {
            'event_type': 'fake_event',
            'payload': {'instance_id': '12345'},
            'publisher': 'fake_publisher',
            'timestamp': '',
            'context': {'project_id': 'fake_project', 'user_id': 'fake_user'}
        }

        event = converter.convert('fake_event', notification)

        self.assertDictEqual(
            {
                'service': 'fake_publisher',
                'project_id': 'fake_project',
                'user_id': 'fake_user',
                'timestamp': ''
            },
            event
        )



# Copyright 2018 OpenStack Foundation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Add last_heartbeat to action execution

Revision ID: 027
Revises: 026
Create Date: 2018-09-05 16:49:50.342349

"""

# revision identifiers, used by Alembic.
revision = '027'
down_revision = '026'

from alembic import op
import datetime
from mistral_lib import utils
from oslo_config import cfg
from sqlalchemy import Column, DateTime, Boolean

CONF = cfg.CONF


def upgrade():
    op.add_column(
        'action_executions_v2',
        Column(
            'last_heartbeat',
            DateTime,
            default=lambda: utils.utc_now_sec() + datetime.timedelta(
                seconds=CONF.action_heartbeat.first_heartbeat_timeout
            )
        )
    )
    op.add_column(
        'action_executions_v2',
        Column('is_sync', Boolean, default=None, nullable=True)
    )



# Copyright (c) 2016 Intel Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from mistral import exceptions as exc
from mistral.tests.unit.rpc.kombu import base
from mistral.tests.unit.rpc.kombu import fake_kombu
from unittest import mock

import queue

with mock.patch.dict('sys.modules', kombu=fake_kombu):
    from mistral.rpc.kombu import base as kombu_base
    from mistral.rpc.kombu import kombu_client


class TestException(exc.MistralException):
    pass


class KombuClientTest(base.KombuTestCase):

    _RESPONSE = "response"

    def setUp(self):
        super(KombuClientTest, self).setUp()

        conf = mock.MagicMock()

        listener_class = kombu_client.kombu_listener.KombuRPCListener

        kombu_client.kombu_listener.KombuRPCListener = mock.MagicMock()

        def restore_listener():
            kombu_client.kombu_listener.KombuRPCListener = listener_class

        self.addCleanup(restore_listener)

        self.client = kombu_client.KombuRPCClient(conf)
        self.ctx = type(
            'context',
            (object,),
            {'to_dict': lambda self: {}}
        )()

    def test_sync_call_result_get(self):
        self.client._listener.get_result = mock.MagicMock(
            return_value={
                kombu_base.TYPE: None,
                kombu_base.RESULT: self.client._serialize_message({
                    'body': self._RESPONSE
                })
            }
        )

        response = self.client.sync_call(self.ctx, 'method')

        self.assertEqual(response, self._RESPONSE)

    def test_sync_call_result_not_get(self):
        self.client._listener.get_result = mock.MagicMock(
            side_effect=queue.Empty
        )

        self.assertRaises(
            exc.MistralException,
            self.client.sync_call,
            self.ctx,
            'method_not_found'
        )

    def test_sync_call_result_type_error(self):
        def side_effect(*args, **kwargs):
            return {
                kombu_base.TYPE: 'error',
                kombu_base.RESULT: TestException()
            }

        self.client._wait_for_result = mock.MagicMock(side_effect=side_effect)

        self.assertRaises(
            TestException,
            self.client.sync_call,
            self.ctx,
            'method'
        )

    def test_async_call(self):
        self.assertIsNone(self.client.async_call(self.ctx, 'method'))



# Copyright 2013 - Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import datetime
import json

from oslo_config import cfg
from oslo_log import log as logging
from oslo_service import periodic_task
from oslo_service import threadgroup

from mistral import context as auth_ctx
from mistral.db.v2 import api as db_api_v2
from mistral import exceptions as exc
from mistral.rpc import clients as rpc
from mistral.services import security
from mistral.services import triggers

LOG = logging.getLogger(__name__)

CONF = cfg.CONF

# {periodic_task: thread_group}
_periodic_tasks = {}


def process_cron_triggers_v2(self, ctx):
    LOG.debug("Processing cron triggers...")

    for trigger in triggers.get_next_cron_triggers():
        LOG.debug("Processing cron trigger: %s", trigger)

        try:
            # Setup admin context before schedule triggers.
            ctx = security.create_context(
                trigger.trust_id,
                trigger.project_id
            )

            auth_ctx.set_ctx(ctx)

            LOG.debug("Cron trigger security context: %s", ctx)

            # Try to advance the cron trigger next_execution_time and
            # remaining_executions if relevant.
            modified = advance_cron_trigger(trigger)

            # If cron trigger was not already modified by another engine.
            if modified:
                LOG.debug(
                    "Starting workflow '%s' by cron trigger '%s'",
                    trigger.workflow.name,
                    trigger.name
                )

                description = {
                    "description": (
                        "Workflow execution created by cron"
                        " trigger '(%s)'." % trigger.id
                    ),
                    "triggered_by": {
                        "type": "cron_trigger",
                        "id": trigger.id,
                    }
                }

                rpc.get_engine_client().start_workflow(
                    trigger.workflow.name,
                    trigger.workflow.namespace,
                    None,
                    trigger.workflow_input,
                    description=json.dumps(description),
                    **trigger.workflow_params
                )
        except Exception:
            # Log and continue to next cron trigger.
            LOG.exception(
                "Failed to process cron trigger %s",
                str(trigger)
            )
        finally:
            auth_ctx.set_ctx(None)


class MistralPeriodicTasks(periodic_task.PeriodicTasks):

    def __init__(self, conf):
        super(MistralPeriodicTasks, self).__init__(conf)

        periodic_task_ = periodic_task.periodic_task(
            spacing=CONF.cron_trigger.execution_interval,
            run_immediately=True,
        )
        self.add_periodic_task(periodic_task_(process_cron_triggers_v2))


def advance_cron_trigger(t):
    modified_count = 0

    try:
        # If the cron trigger is defined with limited execution count.
        if t.remaining_executions is not None and t.remaining_executions > 0:
            t.remaining_executions -= 1

        # If this is the last execution.
        if t.remaining_executions == 0:
            modified_count = triggers.delete_cron_trigger(
                t.name,
                trust_id=t.trust_id,
                delete_trust=False
            )
        else:  # if remaining execution = None or > 0.
            # In case the we are lagging or if the api stopped for some time
            # we use the max of the current time or the next scheduled time.
            next_time = triggers.get_next_execution_time(
                t.pattern,
                max(datetime.datetime.utcnow(), t.next_execution_time)
            )

            # Update the cron trigger with next execution details
            # only if it wasn't already updated by a different process.
            updated, modified_count = db_api_v2.update_cron_trigger(
                t.name,
                {
                    'next_execution_time': next_time,
                    'remaining_executions': t.remaining_executions
                },
                query_filter={
                    'next_execution_time': t.next_execution_time
                }
            )
    except exc.DBEntityNotFoundError as e:
        # Cron trigger was probably already deleted by a different process.
        LOG.debug(
            "Cron trigger named '%s' does not exist anymore: %s",
            t.name, str(e)
        )

    # Return True if this engine was able to modify the cron trigger in DB.
    return modified_count > 0


def setup():
    tg = threadgroup.ThreadGroup()
    pt = MistralPeriodicTasks(CONF)

    ctx = auth_ctx.MistralContext(
        user_id=None,
        project_id=None,
        auth_token=None,
        is_admin=True
    )

    tg.add_dynamic_timer(
        pt.run_periodic_tasks,
        initial_delay=None,
        periodic_interval_max=1,
        context=ctx
    )

    _periodic_tasks[pt] = tg

    return tg


def stop_all_periodic_tasks():
    for tg in _periodic_tasks.values():
        tg.stop()

    _periodic_tasks.clear()



# Copyright 2013 - Mirantis, Inc.
# Copyright 2016 - Brocade Communications Systems, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import cachetools

from oslo_config import cfg
from oslo_db import options
from oslo_db.sqlalchemy import enginefacade
import osprofiler.sqlalchemy
import sqlalchemy as sa
from sqlalchemy.sql import column

from mistral.db.sqlalchemy import sqlite_lock
from mistral import exceptions as exc
from mistral_lib import utils


# Note(dzimine): sqlite only works for basic testing.
options.set_defaults(cfg.CONF, connection="sqlite:///mistral.sqlite")

_DB_SESSION_THREAD_LOCAL_NAME = "__db_sql_alchemy_session__"
_TX_SCOPED_CACHE_THREAD_LOCAL_NAME = "__tx_scoped_cache__"

_facade = None
_sqlalchemy_create_engine_orig = sa.create_engine


def _get_facade():
    global _facade

    if not _facade:
        _facade = enginefacade.LegacyEngineFacade(
            cfg.CONF.database.connection,
            sqlite_fk=True,
            autocommit=False,
            **dict(cfg.CONF.database.items())
        )

        if cfg.CONF.profiler.enabled:
            if cfg.CONF.profiler.trace_sqlalchemy:
                osprofiler.sqlalchemy.add_tracing(
                    sa,
                    _facade.get_engine(),
                    'db'
                )

    return _facade


# Monkey-patching sqlalchemy to set the isolation_level
# as this configuration is not exposed by oslo_db.
def _sqlalchemy_create_engine_wrapper(*args, **kwargs):
    # sqlite (used for unit testing and not allowed for production)
    # does not support READ_COMMITTED.
    # Checking the drivername using the args and not the get_driver_name()
    # method because that method requires a session.
    if args[0].drivername != 'sqlite':
        kwargs["isolation_level"] = "READ_COMMITTED"

    return _sqlalchemy_create_engine_orig(*args, **kwargs)


def get_engine():
    # If the patch was not applied yet.
    if sa.create_engine != _sqlalchemy_create_engine_wrapper:
        # Replace the original create_engine with our wrapper.
        sa.create_engine = _sqlalchemy_create_engine_wrapper

    return _get_facade().get_engine()


def _get_session():
    return _get_facade().get_session()


def _get_thread_local_session():
    return utils.get_thread_local(_DB_SESSION_THREAD_LOCAL_NAME)


def get_tx_scoped_cache():
    return utils.get_thread_local(_TX_SCOPED_CACHE_THREAD_LOCAL_NAME)


def _get_or_create_thread_local_session():
    ses = _get_thread_local_session()

    if ses:
        return ses, False

    ses = _get_session()
    _set_thread_local_session(ses)

    return ses, True


def _set_thread_local_session(session):
    utils.set_thread_local(_DB_SESSION_THREAD_LOCAL_NAME, session)

    if session is not None:
        utils.set_thread_local(
            _TX_SCOPED_CACHE_THREAD_LOCAL_NAME,
            cachetools.LRUCache(maxsize=1000)
        )
    else:
        utils.set_thread_local(_TX_SCOPED_CACHE_THREAD_LOCAL_NAME, None)


def session_aware(param_name="session"):
    """Decorator for methods working within db session."""

    def _decorator(func):
        def _within_session(*args, **kw):
            # If 'created' flag is True it means that the transaction is
            # demarcated explicitly outside this module.
            ses, created = _get_or_create_thread_local_session()

            try:
                kw[param_name] = ses

                result = func(*args, **kw)

                if created:
                    ses.commit()

                return result
            except Exception:
                if created:
                    ses.rollback()
                raise
            finally:
                if created:
                    _set_thread_local_session(None)
                    ses.close()

        _within_session.__doc__ = func.__doc__

        return _within_session

    return _decorator


# Transaction management.


def start_tx():
    """Starts transaction.

    Opens new database session and starts new transaction assuming
    there wasn't any opened sessions within the same thread.
    """
    if _get_thread_local_session():
        raise exc.DataAccessException(
            "Database transaction has already been started."
        )

    _set_thread_local_session(_get_session())


def release_locks_if_sqlite(session):
    if get_driver_name() == 'sqlite':
        sqlite_lock.release_locks(session)


def commit_tx():
    """Commits previously started database transaction."""
    ses = _get_thread_local_session()

    if not ses:
        raise exc.DataAccessException(
            "Nothing to commit. Database transaction"
            " has not been previously started."
        )

    ses.commit()


def rollback_tx():
    """Rolls back previously started database transaction."""
    ses = _get_thread_local_session()

    if not ses:
        raise exc.DataAccessException(
            "Nothing to roll back. Database transaction has not been started."
        )

    ses.rollback()


def end_tx():
    """Ends transaction.

    Ends current database transaction.
    It rolls back all uncommitted changes and closes database session.
    """
    ses = _get_thread_local_session()

    if not ses:
        raise exc.DataAccessException(
            "Database transaction has not been started."
        )

    if ses.dirty:
        rollback_tx()

    release_locks_if_sqlite(ses)

    ses.close()
    _set_thread_local_session(None)


@session_aware()
def get_driver_name(session=None):
    return session.bind.url.drivername


@session_aware()
def get_dialect_name(session=None):
    return session.bind.url.get_dialect().name


@session_aware()
def model_query(model, columns=(), session=None):
    """Query helper.

    :param model: Base model to query.
    :param columns: Optional. Which columns to be queried.
    """

    if columns:
        _columns = (column(c) if isinstance(c, str) else c for c in columns)
        return session.query(*_columns)

    return session.query(model)



# Copyright 2014 - Mirantis, Inc.
# Copyright 2015 Huawei Technologies Co., Ltd.
# Copyright 2020 Nokia Software.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import functools
from oslo_log import log as logging
import pecan
from pecan import hooks
from pecan import rest
from wsme import types as wtypes
import wsmeext.pecan as wsme_pecan

from mistral.api import access_control as acl
from mistral.api.controllers.v2 import resources
from mistral.api.controllers.v2 import types
from mistral.api.controllers.v2 import validation
from mistral.api.hooks import content_type as ct_hook
from mistral import context
from mistral.db.v2 import api as db_api
from mistral import exceptions as exc
from mistral.lang import parser as spec_parser
from mistral.services import actions as action_service
from mistral.services import adhoc_actions
from mistral.utils import filter_utils
from mistral.utils import rest_utils

from mistral_lib import utils


LOG = logging.getLogger(__name__)


def _action_descriptor_to_resource(action_desc):
    return resources.Action(
        id=getattr(action_desc, 'id', action_desc.name),
        name=action_desc.name,
        description=action_desc.description,
        input=action_desc.params_spec,
        namespace=action_desc.namespace,
        project_id=action_desc.project_id,
        scope=action_desc.scope,
        definition=getattr(action_desc, 'definition', ''),
        is_system=False,
        tags=getattr(action_desc, 'tags', None),
        created_at=utils.datetime_to_str(
            getattr(action_desc, 'created_at', '')
        ),
        updated_at=utils.datetime_to_str(
            getattr(action_desc, 'updated_at', '')
        )
    )


class ActionsController(rest.RestController, hooks.HookController):
    # TODO(nmakhotkin): Have a discussion with pecan/WSME folks in order
    # to have requests and response of different content types. Then
    # delete ContentTypeHook.
    __hooks__ = [ct_hook.ContentTypeHook("application/json", ['POST', 'PUT'])]

    validate = validation.SpecValidationController(
        spec_parser.get_action_list_spec_from_yaml)

    @rest_utils.wrap_wsme_controller_exception
    @wsme_pecan.wsexpose(resources.Action, wtypes.text, wtypes.text)
    def get(self, identifier, namespace=''):
        """Return the named action.

        :param identifier: ID or name of the Action to get.
        :param namespace: The namespace of the action.
        :param fields: Optional. A specified list of fields of the resource to
                       be returned. 'id' will be included automatically in
                       fields if it's not provided.
        """

        acl.enforce('actions:get', context.ctx())

        LOG.debug("Fetch action [identifier=%s]", identifier)

        action_provider = action_service.get_system_action_provider()

        # Here we assume that the action search might involve DB operations
        # so we need to apply the regular retrying logic as everywhere else.
        action_desc = rest_utils.rest_retry_on_db_error(
            action_provider.find
        )(identifier, namespace=namespace)

        if action_desc is None:
            # TODO(rakhmerov): We need to change exception class so that
            # it's not DB specific. But it should be associated with the
            # same HTTP code.
            raise exc.DBEntityNotFoundError(
                'Action not found [name=%s, namespace=%s]'
                % (identifier, namespace)
            )

        return _action_descriptor_to_resource(action_desc)

    @rest_utils.wrap_pecan_controller_exception
    @pecan.expose(content_type="text/plain")
    def put(self, identifier=None, namespace=''):
        """Update one or more actions.

        :param identifier: Optional. If provided, it's UUID or name of an
            action. Only one action can be updated with identifier param.
        :param namespace: Optional. If provided, it's the namespace that
            the action is under.

        NOTE: This text is allowed to have definitions
            of multiple actions. In this case they all will be updated.
        """
        acl.enforce('actions:update', context.ctx())

        definition = pecan.request.text

        LOG.debug("Update action(s) [definition=%s]", definition)

        namespace = namespace or ''

        scope = pecan.request.GET.get('scope', 'private')

        resources.Action.validate_scope(scope)

        if scope == 'public':
            acl.enforce('actions:publicize', context.ctx())

        @rest_utils.rest_retry_on_db_error
        def _update_actions():
            with db_api.transaction():
                return adhoc_actions.update_actions(
                    definition,
                    scope=scope,
                    identifier=identifier,
                    namespace=namespace
                )

        db_acts = _update_actions()

        action_list = [
            resources.Action.from_db_model(db_act) for db_act in db_acts
        ]

        return resources.Actions(actions=action_list).to_json()

    @rest_utils.wrap_pecan_controller_exception
    @pecan.expose(content_type="text/plain")
    def post(self, namespace=''):
        """Create a new action.

        :param namespace: Optional. The namespace to create the ad-hoc action
            in. actions with the same name can be added to a given
            project if they are in two different namespaces.
            (default namespace is '')

        NOTE: This text is allowed to have definitions
            of multiple actions. In this case they all will be created.
        """
        acl.enforce('actions:create', context.ctx())

        namespace = namespace or ''

        definition = pecan.request.text
        scope = pecan.request.GET.get('scope', 'private')
        pecan.response.status = 201

        resources.Action.validate_scope(scope)

        if scope == 'public':
            acl.enforce('actions:publicize', context.ctx())

        LOG.debug("Create action(s) [definition=%s]", definition)

        @rest_utils.rest_retry_on_db_error
        def _create_action_definitions():
            with db_api.transaction():
                return adhoc_actions.create_actions(
                    definition,
                    scope=scope,
                    namespace=namespace
                )

        db_acts = _create_action_definitions()

        action_list = [
            resources.Action.from_db_model(db_act) for db_act in db_acts
        ]

        return resources.Actions(actions=action_list).to_json()

    @rest_utils.wrap_wsme_controller_exception
    @wsme_pecan.wsexpose(None, wtypes.text, wtypes.text, status_code=204)
    def delete(self, identifier, namespace=''):
        """Delete the named action.

        :param identifier: Name or UUID of the action to delete.
        :param namespace: The namespace of which the action is in.
        """
        acl.enforce('actions:delete', context.ctx())

        LOG.debug("Delete action [identifier=%s]", identifier)

        @rest_utils.rest_retry_on_db_error
        def _delete_action_definition():
            with db_api.transaction():
                db_model = db_api.get_action_definition(
                    identifier,
                    namespace=namespace
                )

                if db_model.is_system:
                    raise exc.DataAccessException(
                        "Attempt to delete a system action: %s" % identifier
                    )

                db_api.delete_action_definition(
                    identifier,
                    namespace=namespace
                )

        _delete_action_definition()

    @rest_utils.wrap_wsme_controller_exception
    @wsme_pecan.wsexpose(resources.Actions, wtypes.text, int, types.uniquelist,
                         types.list, types.uniquelist, wtypes.text,
                         wtypes.text, resources.SCOPE_TYPES, wtypes.text,
                         wtypes.text, wtypes.text, wtypes.text,
                         wtypes.text, wtypes.text, wtypes.text)
    def get_all(self, marker=None, limit=None, sort_keys='name',
                sort_dirs='asc', fields='', created_at=None,
                name=None, scope=None, tags=None,
                updated_at=None, description=None, definition=None,
                is_system=None, input=None, namespace=''):
        """Return all actions.

        :param marker: Optional. Pagination marker for large data sets.
        :param limit: Optional. Maximum number of resources to return in a
                      single result. Default value is None for backward
                      compatibility.
        :param sort_keys: Optional. Columns to sort results by.
                          Default: name.
        :param sort_dirs: Optional. Directions to sort corresponding to
                          sort_keys, "asc" or "desc" can be chosen.
                          Default: asc.
        :param fields: Optional. A specified list of fields of the resource to
                       be returned. 'id' will be included automatically in
                       fields if it's not provided, since it will be used when
                       constructing 'next' link.
        :param name: Optional. Keep only resources with a specific name.
        :param scope: Optional. Keep only resources with a specific scope.
        :param definition: Optional. Keep only resources with a specific
                           definition.
        :param is_system: Optional. Keep only system actions or ad-hoc
                          actions (if False).
        :param input: Optional. Keep only resources with a specific input.
        :param description: Optional. Keep only resources with a specific
                            description.
        :param tags: Optional. Keep only resources containing specific tags.
        :param created_at: Optional. Keep only resources created at a specific
                           time and date.
        :param updated_at: Optional. Keep only resources with specific latest
                           update time and date.
        :param namespace: Optional. The namespace of the action.
        """
        acl.enforce('actions:list', context.ctx())

        filters = filter_utils.create_filters_from_request_params(
            created_at=created_at,
            name=name,
            scope=scope,
            tags=tags,
            updated_at=updated_at,
            description=description,
            definition=definition,
            is_system=is_system,
            input=input,
            namespace=namespace
        )

        LOG.debug(
            "Fetch actions. marker=%s, limit=%s, sort_keys=%s, "
            "sort_dirs=%s, filters=%s",
            marker,
            limit,
            sort_keys,
            sort_dirs,
            filters
        )

        sort_keys = ['name'] if sort_keys is None else sort_keys
        sort_dirs = ['asc'] if sort_dirs is None else sort_dirs
        fields = [] if fields is None else fields

        if fields and 'name' not in fields:
            fields.insert(0, 'name')

        rest_utils.validate_query_params(limit, sort_keys, sort_dirs)

        action_provider = action_service.get_system_action_provider()

        # Here we assume that the action search might involve DB operations
        # so we need to apply the regular retrying logic as everywhere else.
        action_descriptors = rest_utils.rest_retry_on_db_error(
            action_provider.find_all
        )(
            namespace=namespace,
            limit=limit,
            sort_fields=sort_keys,
            sort_dirs=sort_dirs,
            filters=filters
        )

        # We can't guarantee that at this point the collection of action
        # descriptors is properly filtered and sorted.

        # Apply filters.
        action_descriptors = filter(
            lambda a_d: filter_utils.match_filters(a_d, filters),
            action_descriptors
        )

        # Apply sorting.
        def action_descriptor_sort(a_ds, keys, dirs):
            def compare_(a_d1, a_d2):
                for key, dir in zip(keys, dirs):
                    a_d1 = getattr(a_d1, key, None)
                    a_d2 = getattr(a_d2, key, None)

                    if a_d1 is None and a_d2 is None:
                        ret = 0
                    elif a_d1 is None and a_d2 is not None:
                        ret = -1
                    elif a_d1 is not None and a_d2 is None:
                        ret = 1
                    else:
                        if type(a_d1) is type(a_d2):
                            ret = (a_d1 > a_d2) - (a_d1 < a_d2)
                        else:
                            ret = 1
                    if ret:
                        return ret * (1 if dir == 'asc' else -1)
                return 0
            return sorted(a_ds, key=functools.cmp_to_key(compare_))

        action_descriptors = action_descriptor_sort(action_descriptors,
                                                    sort_keys, sort_dirs)
        start = 0
        for i, a_d in enumerate(action_descriptors):
            if a_d.name == marker:
                start = i
                break

        if limit and limit > 0:
            end = start + limit
            action_descriptors = action_descriptors[start:end]

        action_resources = [
            _action_descriptor_to_resource(a_d)
            for a_d in action_descriptors
        ]

        # TODO(rakhmerov): Fix pagination so that it doesn't work with
        # the 'id' field as a marker. We can't use IDs anymore. "name"
        # seems a good candidate for this.
        return resources.Actions.convert_with_links(
            action_resources,
            limit,
            pecan.request.application_url,
            sort_keys=','.join(sort_keys),
            sort_dirs=','.join(sort_dirs),
            **filters
        )



# Copyright 2015 - Mirantis, Inc.
# Copyright 2016 - Brocade Communications Systems, Inc.
# Copyright 2020 Nokia Software.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from oslo_log import log as logging
from osprofiler import profiler
import traceback as tb

from mistral.db.v2.sqlalchemy import models
from mistral.engine import actions
from mistral.engine import task_handler
from mistral import exceptions as exc
from mistral.services import actions as action_service


LOG = logging.getLogger(__name__)


@profiler.trace('action-handler-on-action-complete', hide_args=True)
def on_action_complete(action_ex, result):
    task_ex = action_ex.task_execution

    action = _build_action(action_ex)

    try:
        action.complete(result)
    except exc.MistralException as e:
        msg = (
            "Failed to complete action [error=%s, action=%s, task=%s]:\n%s"
            % (e, action_ex.name, task_ex.name, tb.format_exc())
        )

        LOG.error(msg)

        action.fail(msg)

        if task_ex:
            task_handler.force_fail_task(task_ex, msg)

        return

    if task_ex:
        task_handler.schedule_on_action_complete(action_ex)


@profiler.trace('action-handler-on-action-update', hide_args=True)
def on_action_update(action_ex, state):
    task_ex = action_ex.task_execution

    action = _build_action(action_ex)

    try:
        action.update(state)
    except exc.MistralException as e:
        # If the update of the action execution fails, do not fail
        # the action execution. Log the exception and re-raise the
        # exception.
        msg = (
            "Failed to update action [error=%s, action=%s, task=%s]:\n%s"
            % (e, action_ex.name, task_ex.name, tb.format_exc())
        )

        LOG.error(msg)

        raise

    if task_ex:
        task_handler.schedule_on_action_update(action_ex)


@profiler.trace('action-handler-build-action', hide_args=True)
def _build_action(action_ex):
    if isinstance(action_ex, models.WorkflowExecution):
        return actions.WorkflowAction(
            wf_name=action_ex.name,
            action_ex=action_ex
        )

    action_desc = action_service.get_system_action_provider().find(
        action_ex.name,
        action_ex.workflow_namespace
    )

    if action_desc is None:
        raise exc.InvalidActionException(
            "Failed to find action [action_name=%s]" %
            action_ex.name
        )

    return actions.RegularAction(action_desc, action_ex)


def build_action_by_name(action_name, namespace=''):
    action_desc = action_service.get_system_action_provider().find(
        action_name,
        namespace=namespace
    )

    if action_desc is None:
        raise exc.InvalidActionException(
            "Failed to find action [action_name=%s]" %
            action_name
        )

    return actions.RegularAction(action_desc)



# Copyright 2020 Nokia Software.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.


from oslo_config import cfg
import types

from mistral_lib import actions as ml_actions
from mistral_lib import serialization
from mistral_lib.utils import inspect_utils

from mistral.db.v2 import api as db_api

CONF = cfg.CONF


class DynamicAction(ml_actions.Action):
    def __init__(self, action, code_source_id, namespace=''):
        super(DynamicAction, self).__init__()

        self.action = action
        self.namespace = namespace
        self.code_source_id = code_source_id

    @classmethod
    def get_serialization_key(cls):
        return '%s.%s' % (DynamicAction.__module__, DynamicAction.__name__)

    def run(self, context):
        return self.action.run(context)

    def is_sync(self):
        return self.action.is_sync()


class DynamicActionDescriptor(ml_actions.PythonActionDescriptor):
    def __init__(self, name, cls_name, action_cls, code_source_id, version,
                 action_cls_attrs=None, namespace='', project_id=None,
                 scope=None):
        super(DynamicActionDescriptor, self).__init__(
            name,
            action_cls,
            action_cls_attrs,
            namespace,
            project_id,
            scope
        )

        self.cls_name = cls_name
        self.code_source_id = code_source_id
        self.version = version

    def __repr__(self):
        return 'Dynamic action [name=%s, cls=%s , code_source_id=%s,' \
               ' version=%s]' % (
                   self.name,
                   self._action_cls,
                   self.code_source_id,
                   self.version
               )

    def instantiate(self, params, wf_ctx):
        if not self._action_cls_attrs:
            # No need to create new dynamic type.
            return DynamicAction(
                self._action_cls(**params),
                self.code_source_id,
                self.namespace
            )

        dynamic_cls = type(
            self._action_cls.__name__,
            (self._action_cls,),
            **self._action_cls_attrs
        )

        return DynamicAction(
            dynamic_cls(**params),
            self.code_source_id,
            self.namespace
        )


class DynamicActionSerializer(serialization.DictBasedSerializer):
    def serialize_to_dict(self, entity):
        cls = type(entity.action)

        return {
            'cls_name': cls.__name__,
            'cls_attrs': inspect_utils.get_public_fields(cls),
            'data': vars(entity.action),
            'code_source_id': entity.code_source_id,
            'namespace': entity.namespace,
        }

    def deserialize_from_dict(self, entity_dict):
        cls_name = entity_dict['cls_name']

        mod = _get_python_module(
            entity_dict['code_source_id'],
            entity_dict['namespace']
        )

        cls = getattr(mod[0], cls_name)

        cls_attrs = entity_dict['cls_attrs']

        if cls_attrs:
            cls = type(cls.__name__, (cls,), cls_attrs)

        action = cls.__new__(cls)

        for k, v in entity_dict['data'].items():
            setattr(action, k, v)

        return DynamicAction(
            action,
            entity_dict['code_source_id'],
            entity_dict['namespace']
        )


def _get_python_module(code_source_id, namespace=''):
    code_source = db_api.get_code_source(
        code_source_id,
        namespace=namespace
    )

    mod = _load_python_module(code_source.name, code_source.content)

    return mod, code_source.version


def _load_python_module(fullname, content):
    mod = types.ModuleType(fullname)

    exec(content, mod.__dict__)

    return mod


serialization.register_serializer(DynamicAction, DynamicActionSerializer())


class DynamicActionProvider(ml_actions.ActionProvider):
    """Provides dynamic actions."""

    def __init__(self, name='dynamic'):
        super().__init__(name)

        # {code_source_id => (python module, version)}
        self._code_sources = dict()

    def ensure_latest_module_version(self, action_def):
        # We need to compare the version of the corresponding module
        # that's already loaded into memory with the version stored in
        # DB and reimport the module if the DB version is higher.

        code_src_id = action_def.code_source_id

        # TODO(rakhmerov): To avoid this DB call we need to store code source
        # versions also in the dynamic action definition model.
        db_ver = db_api.get_code_source(code_src_id, fields=['version'])[0]

        if db_ver > self._code_sources.get(code_src_id, (None, -1))[1]:
            # Reload module.
            code_src = db_api.get_code_source(code_src_id)

            module = _load_python_module(code_src.name, code_src.content)

            self._code_sources[code_src_id] = (module, code_src.version)
        else:
            module = self._code_sources[code_src_id][0]

        return module

    def _get_action_class(self, action_def):
        module = self.ensure_latest_module_version(action_def)

        return getattr(module, action_def.class_name)

    def _build_action_descriptor(self, action_def):
        action_cls = self._get_action_class(action_def)

        return DynamicActionDescriptor(
            name=action_def.name,
            cls_name=action_def.class_name,
            action_cls=action_cls,
            code_source_id=action_def.code_source_id,
            version=1,
            project_id=action_def.project_id,
            scope=action_def.scope
        )

    def find(self, action_name, namespace=None):
        action_def = db_api.load_dynamic_action_definition(
            action_name,
            namespace
        )

        if action_def is None:
            return None

        return self._build_action_descriptor(action_def)

    def find_all(self, namespace='', limit=None, sort_fields=None,
                 sort_dirs=None, filters=None):
        if filters is None:
            filters = dict()

        filters['namespace'] = {'eq': namespace}

        action_defs = db_api.get_dynamic_action_definitions(
            limit=limit,
            sort_keys=sort_fields,
            sort_dirs=sort_dirs,
            **filters
        )

        return [self._build_action_descriptor(a_d) for a_d in action_defs]
